; Skype scripts that can run even when Skype is not visible.
;
; Author:  Doug Lee
/*
Copyright (c) 2006-2017 Doug Lee.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* The names of the copyright holders and contributors may not be used to
  endorse or promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

use "skype_compat.jsb"  ; compatibility with older JAWS versions

;#pragma StringComparison partial
;#pragma usePoFile 0
include "hjconst.jsh"
include "common.jsm"
include "skypewatch.jsh"
include "skypewatch.jsm"
include "MSAAConst.jsh"
const
; Values for SkypeWatch___state (the state of Skype itself)
	_StateOff = 0,
	_StateOn = 1,
; The maximum number of SkypeWatch messages to save for retrieval by SkypeWatch__getAlert().
	_AlertCountMax = 10

globals
; The state of Skype itself.
	int SkypeWatch___state,
; The online/offline/etc. state of the current Skype user.
	string SkypeWatch___status,
; The saved alerts for SkypeWatch__getAlert().
	string SkypeWatch___alerts,
; The windows being watched for alerts; see SkypeWatch___alertAddWindow() for structure.
	collection SkypeWatch___alertWindows,
; The getTickCount() value at last windowCreatedEvent call on a Skype alert window.
; Used to minimize scans for windows missed when that event doesn't fire when expected.
	int SkypeWatch___WCETick

void function autoStartEvent()
SkypeWatch___alertWindows = new collection
SkypeWatch___WCETick = 0
; Default scripts have to set a log file name, or their logs will go to different files depending on which app is in focus.
; This makes the SkypeWatch scripts use the same log file as the Skype scripts.
SkypeWatch__log__open("Skype.log", False)
SkypeWatch_WCECalled = False
SkypeWatch___cleanIni()
SkypeWatch__setKeys()
SkypeWatch___loadAlerts()
SkypeWatch__supdate__checkFiles("|default")
scheduleFunction("SkypeWatch___alertWatcher", 10)
scheduleFunction("SkypeWatch__SkypeWatcher", 20)
endFunction

void function SkypeWatch___cleanIni()
; Clean up skype.ini from problems caused by CR and/or LF characters in alert text.
; This problem was noted by Michele Barbi on 2016-08-25.
var string fname = sutil__INIFileName
var int cnt = 0
var int safety
for safety = 1 to 25
	var int cnt1 = SkypeWatch___cleanIni0(fname)
	if !cnt1
		if cnt
			sayMessage(OT_Error, formatString(_msgIniCleanCount, intToString(cnt), fname))
		endIf
		return
	endIf
	cnt = cnt +cnt1
endFor
if cnt
	sayMessage(OT_Error, formatString(_msgIniCleanCount, intToString(cnt), fname))
endIf
endFunction

int function SkypeWatch___cleanIni0(string fname)
; Clean up skype.ini from problems caused by CR and/or LF characters in alert text.
; Helper for SkypeWatch___cleanIni(). Running this process just once can miss sections.
var string sects = iniReadSectionNames(fname)
if !sects
	return
endIf
var string sep = "|"
var int n = stringSegmentCount(sects, sep)
var int nremoved = 0
while n
	var string sect = stringSegment(sects, sep, n)
	if stringCompare(sect, "Toggles", False) != 0
		if iniRemoveSection(sect, fname)
		&& iniFlush(fname)
			nremoved = nremoved +1
		else
			sayMessage(OT_Error, formatString(_msgIniRemoveSectFail, sect, fname))
		endIf
	endIf
	n = n -1
endWhile
return nremoved
endFunction

void function SkypeWatch___say(string s, int always)
if always || SkypeWatch__opt__getUserOption(opt__SV_SkypeWatchMessages)
	sayUsingVoice(VCTX_Message, formatString(sutil__msgSkypeWatchAlertFormat, s), OT_Status)
endIf
endFunction

void function SkypeWatch__setKey(string sect, string keyName, string val)
; Assign the given default key as indicated in a way that works in any of several JAWS versions.
var
	string fname,
	string curVal
if !stringCompare(getJAWSUserName(), "system", False)
	; Avoid error messages in JAWS 10+ with security patches from 2009:
	; When running as System, JAWS apparently can't handle ini file lookups and other filesystem accesses properly.
	return
endIf
fname = "default.jkm"
fname = getJAWSSettingsDirectory() +"\\" +fname
curVal = iniReadString(sect, keyName, "", fname)
if !stringCompare(curVal, val, False)
	return
endIf
if !stringIsBlank(curVal)
	SkypeWatch___say(formatString(msgRLMKeyChanging, keyName, curVal, val), True)
endIf
if !iniWriteString(sect, keyName, val, fname)
	SkypeWatch___say(formatString(msgRLMKeyChangeFailed, keyName, val), True)
endIf
endFunction

void function SkypeWatch__setKeys()
; Set up any default keystrokes for SkypeWatch.
SkypeWatch__setKey("Common Keys", "alt+shift+f9", "swSayAlert")
SkypeWatch__setKey("Common Keys", "Control+ScrollLock+F9", "swSayAlert")
endFunction

void function SkypeWatch___loadAlerts()
; Load alerts from the .ini file to carry them across JAWS restarts.
if SkypeWatch__xlist__itemCount(SkypeWatch___alerts) > 0
	return
endIf
SkypeWatch___alerts = SkypeWatch__xlist__fromJAWSList(iniReadString(sutil__INISectName, sutil__IniSWAKey, "", sutil__INIFileName), sutil__IniSWASep)
if !SkypeWatch___alerts
	SkypeWatch___alerts = SkypeWatch__xlist__new()
endIf
endFunction

void function SkypeWatch___saveAlerts()
var string alerts = SkypeWatch___alerts
alerts = stringReplaceSubstrings(alerts, "\13", " ")
alerts = stringReplaceSubstrings(alerts, "\10", " ")
iniWriteString(sutil__INISectName, sutil__IniSWAKey, SkypeWatch__xlist__toJAWSList(alerts, sutil__IniSWASep), sutil__INIFileName)
endFunction

void function SkypeWatch___alertAddWindow(handle hwnd, int offset)
; Add a window to the list to watch. Fields:
;	hwnd: The window handle.
;	tick: The approximate getTickCount() value at window creation time.
;	spoken: Number of times already announced.
; Returns True if the window was added and False if not (which happens when it's already there).
if !SkypeWatch___alertWindows
	SkypeWatch___alertWindows = new collection
endIf
var collection hwnds = SkypeWatch___alertWindows
var string shwnd = intToString(hwnd)
if !collectionItemExists(hwnds, shwnd)
	var collection rec rec = new collection
	hwnds[shwnd] = rec
	rec.hwnd = hwnd
	rec.tick = getTickCount() + offset
	rec.spoken = 0
	return True
endIf
return False
endFunction

string function SkypeWatch___alertGetText(handle hwnd)
; Gets the text of the given alert window for speaking.
var
	object o, int childID, int n,
	int iRole,
	string sText
o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
if !o
	return ""
endIf
sText = o.accValue(0)
if sText
	return sText
endIf
childID = 1
n = o.accChildCount
while childID <= n
	iRole = o.accRole(childID)
	if iRole != Role_System_PushButton
		sText = formatString("%1, %2", sText, o.accName(childID))
		if iRole != Role_System_StaticText
			sText = sText +" " +getRoleText(iRole)
		endIf
	endIf
	childID = childID +1
endWhile
sText = stringChopLeft(sText, 2)  ; ", "
return sText
endFunction

void function SkypeWatch___alertWatcher()
; Watch already-noticed notification windows until they are old enough to scan for text, then process them.
; Also catch new alert windows in case the WindowCreatedEvent isn't working, which happens. [DGL, 2016-04-02]
; Called from AutoStartEvent to kick things off.
; Also called via scheduleFunction from within.
scheduleFunction("SkypeWatch___alertWatcher", 7)
SkypeWatch___addNewAlertWindows()

var string shwnd, collection rec
forEach shwnd in SkypeWatch___alertWindows
	rec = SkypeWatch___alertWindows[shwnd]
	var handle hwnd = rec.hwnd
	var int tick = rec.tick
	; The next line specifies how many milliseconds to wait from window creation to windowText retrieval.
	; The time from window creation to destruction should be five seconds in most cases and (hopefully) never shorter.
	tick = tick +500
	if getTickCount() > tick
		if !getWindowStyleBits(hwnd)
			; The window has disappeared.
			; Tests indicate it is safe to delete this during a forEach loop. [DGL, 2016-04-02]
			collectionRemoveItem(SkypeWatch___alertWindows, shwnd)
		elif rec.spoken
			; Do nothing because we already did.
		else
			var string sText = SkypeWatch___alertGetText(hwnd)
			if stringLength(sText)
				SkypeWatch___alert(sText)
				rec.spoken = rec.spoken +1
			else
				; Window not ready yet, check back real soon.
				scheduleFunction("SkypeWatch___alertWatcher", 2)
				; Continue checking other windows though.
			endIf
		endIf
	endIf
endForEach
endFunction

int function SkypeWatch___addNewAlertWindows()
; Add any alert windows not already added.
; Does not scan for new windows if the WindowCreatedEvent approach seems to be working.
; Returns the number of new windows added.
var int tc = getTickCount()
if tc -SkypeWatch___WCETick < 10000
	return 0
endIf
var handle hwnd = SkypeWatch__appwin__findFirstSkypeNotifyWindow()
if !hwnd
	return 0
endIf
var int cnt = 0
while hwnd
if isWindowVisible(hwnd)
&& SkypeWatch__appwin__isSkypeNotifyWindow(hwnd)
	; Offset should match the largest scheduleFunction delay for scheduling SkypeWatch___alertWatcher(), converted to milliseconds.
	SkypeWatch___alertAddWindow(hwnd, 1000)
	cnt = cnt +1
endIf
	hwnd = getNextWindow(hwnd)
endWhile
	return cnt
endFunction

void function SkypeWatch___alert(string s)
if stringIsBlank(s)
	; Shouldn't happen; see sw_processWindow().
	return
endIf
SkypeWatch___say(s, False)
if !SkypeWatch___alerts
	SkypeWatch___alerts = SkypeWatch__xlist__new()
endIf
s = formatString(msgSWFormat, s, sysGetTime())
SkypeWatch__xlist__insertItem(SkypeWatch___alerts, 1, s)
if SkypeWatch__xlist__itemCount(SkypeWatch___alerts) >= _AlertCountMax
	SkypeWatch__xlist__resize(SkypeWatch___alerts, _AlertCountMax)
endIf
SkypeWatch___saveAlerts()
endFunction

string function SkypeWatch__getAlert(int which)
; Get the requested alert, 0 meaning current and anything else going back in time.
var
	int cnt
if which < 0
	which = 0 -which
endIf
which = which +1
cnt = SkypeWatch__xlist__itemCount(SkypeWatch___alerts)
if which > cnt
	return ""
endIf
return SkypeWatch__xlist__getItem(SkypeWatch___alerts, which)
endFunction

script swSayAlert(int which)
var
	string alert
if !isSameScript()
	if which < 0
		which = which +1
	elif which > 0
		which = which -1
	endIf
	alert = SkypeWatch__getAlert(which)
	if !alert
		if which == 0 || !SkypeWatch__getAlert(0)
			alert = msgNoAlerts
		else
			alert = msgNotThatManyAlerts
		endIf
	endIf
	SkypeWatch___say(alert, True)
	return
endIf
alert = SkypeWatch__opt__userOptionCycler(opt__SV_SkypeWatchMessages, opt__msgSkypeWatchMessageChoices, False)
alert = formatString(msgSWMessagesSpeak, alert)
SkypeWatch___say(alert, True)
endScript

void function SkypeWatch___skypeWatchHelper()
; Does the work of SkypeWatch__SkypeWatcher().
var int tc = getTickCount()

; If neither speech nor Braille are in use, do nothing because announcements wouldn't happen anyway.
; This can happen, for example, when Braille and speech are disabled in a DOS box.
if !speechInUse() && !BrailleInUse()
	return
endIf

; What if Skype isn't running...
if !SkypeWatch__appwin__getSkypeWindow(appwin__SWT_Main)
	if SkypeWatch___state != _StateOff
		SkypeWatch___say(sutil__msgUnloaded, False)
		SkypeWatch___state = _StateOff
		SkypeWatch___status = msgStatUnknown
	endIf
	return
endIf
SkypeWatch___state = _StateOn

; Check the current user's status for changes.
if SkypeWatch__opt__getUserOption(opt__SV_SpeakOnlineStatus)
	var string newstat = SkypeWatch__sutil__getMyOnlineStatus()
	if stringCompare(newstat, SkypeWatch___status, True)
		SkypeWatch___say(formatString(msgStat, newstat), False)
		SkypeWatch___status = newstat
	endIf
endIf
endFunction

void function SkypeWatch__SkypeWatcher()
SkypeWatch_lastTick = getTickCount()  ; makes SkypeWatch__sutil__isSkypeWatchLoaded() work.
SkypeWatch___skypeWatchHelper()
scheduleFunction("SkypeWatch__SkypeWatcher", 20)
endFunction

Void Function WindowCreatedEvent (handle hWindow, int nLeft, int nTop, int nRight, int nBottom)
windowCreatedEvent(hWindow, nLeft, nTop, nRight, nBottom)
; This signals SkypeWatch__sutil__isSkypeWatchLoaded() that this WindowCreatedEvent was called since the last SkypeWatch restart.
; This indicates that SkypeWatch is loaded ahead of default.jsb and can get events.
SkypeWatch_WCECalled = True
hWindow = getTopLevelWindow(hWindow)
if SkypeWatch__appwin__isSkypeNotifyWindow(hWindow)
	; Offset 0 because this window was just created.
	SkypeWatch___alertAddWindow(hWindow, 0)
	SkypeWatch___WCETick = getTickCount()
endIf
endFunction

; Interceptions of JAWS 15 scripts, to prevent "Unknown script call" messages.
script repeatLastSkypeNotification()
; performScript allows no parameters; this works around that.
formatStringWithEmbeddedFunctions("<$swSayAlert(1)>")
endScript


;======================== Module sutil (from SkypeUtil) ========================
; Utility code for the Skype scripts.
;
; Author:  Doug Lee

object function SkypeWatch__sutil__getFocusObject(int byRef childID)
; A GetFocusObject wrapper that accounts for Skype's sometimes returning the wrong thing.
var
	object o,
	variant v
o = getFocusObject(childID)
v = o.accFocus
if !stringLength(v) && v
	o = v
	childID = 0
elif v
	childID = v +0
endIf
return o
endFunction

int function SkypeWatch__sutil__isSkypeWatchLoaded()
; Returns True if SkypeWatch is running:
;	- 1 if SkypeWatch is running via the old (Use "SkypeWatch.jsb" in default.jss) installation method.
;	- 2 if SkypeWatch is running via the new method (default.jss is loader of SkypeWatch and default_fs.jsb).
; The two are distinguished from one another by whether SkypeWatche's WindowCreatedEvent has been called since the last SkypeWatch restart.
; SkypeWatch_lastTick is updated by SkypeWatch::SkypeWatcher(), which runs once every five seconds and once two seconds after SkypeWatch first loads.
; If it hasn't been updated in 15 seconds, SkypeWatch is surely not around to do it.
; For some reason though, the delay has been seen to exceed 8 seconds. [DGL, 2005-02-08]
var
	int diff
diff = getTickCount() -SkypeWatch_lastTick
;sayInteger(diff)
if diff < 15000
	if SkypeWatch_WCECalled
		return 2
	endIf
	return 1
endIf
return 0
endFunction

globals int SkypeWatch__sutil___saidUnsupported
void function SkypeWatch__sutil__sayMessageToUserBuffer(string s, int format)
var
	int ot
s = s +"\13\10\13\10" +cMsgBuffExit
ot = OT_User_Buffer
if format
	sayFormattedMessage(ot, s)
else
	sayMessage(ot, s)
endIf
endFunction

void function SkypeWatch__sutil__sayURMessage(string sLong, string sShort, int format)
; Similar to sayMessageToUserBuffer:  Allows JAWS 4.x/5.x/6.x to do things the same way.
; A message sent this way will flash in Braille in JAWS 6+.
var
	int wasSpeechOff,
	int ot
wasspeechOff = isSpeechOff()
if wasSpeechOff
	speechOn()
endIf
ot = OT_User_Requested_Information
if sShort
	if format
		sayFormattedMessage(ot, sLong, sShort)
	else
		sayMessage(ot, sLong, sShort)
	endIf
else
	if format
		sayFormattedMessage(ot, sLong)
	else
		sayMessage(ot, sLong)
	endIf
endIf
if wasSpeechOff
	speechOff()
endIf
endFunction

int function SkypeWatch__sutil__isMSAASupport(handle hwnd)
; Determines whether the given window provides MSAA support.
var
	object o, int childID, int iRole
o = getObjectFromEvent(hwnd, -4, 0, childID)
if !o
	return False
endIf
iRole = o.accRole(childID)
if iRole == Role_System_Client || iRole == Role_System_Window
	return False
endIf
return True
endFunction

string function SkypeWatch__sutil__getToolBarItemDesc(string givenName, object o, int childId, int includeSeps)
; Construct and return the displayable name/description for the indicated toolbar item.
; GivenName, if non-null, will be used as the name instead of whatever is found on screen or in MSAA.
; As a special case, if givenName contains a slash and the button is a check box,
; the part left of the slash will be used as the name if the box is unchecked, and the right part if checked.
; includeSeps determines whether separators will return something meaningful or a null string.
; Author:  Doug Lee
var
	int iRole, int iState,
	int left, int right, int top, int bottom,
	string name, string states,
	int altnames
iRole = o.accRole(childId)
if iRole == Role_System_Window
	return ""
elif iRole == Role_System_Separator
	if includeSeps
		if givenName
			return givenName
		else
			return sutil__msgSep
		endIf
	else
		return ""
	endIf
endIf
states = ""
name = givenName
if !name
	name = o.accName(childId)
endIf
if !name
	; Example:  WordPad:  No accName but labeling graphics puts names in the object's rectangle.
	if SkypeWatch__sutil__MSAA__rect(o, childId, left, right, top, bottom)
		name = getTextInRect(left, top, right, bottom)
	endIf
endIf
if !name
	; TODO:  When separators are skipped above, item numbers here will skip also.
	name = formatString(sutil__msgUnnamedItem, intToString(childId))
	states = states +" " +sutil__msgUnnamed
endIf
name = name +" " +getRoleText(iRole)
iState = o.accState(childId)
if iState & State_System_Unavailable
	states = states + " " +cMsgNotavailable
endIf
if iState & State_System_OffScreen
	states = states + " " +sutil__msgOffScreen
elif iState & State_System_Invisible
	states = states + " " +sutil__msgInvisible
endIf
if iRole == Role_System_CheckButton || iRole == Role_System_RadioButton
	if iRole == Role_System_CheckButton
		if stringContains(name, "/")
			; When given alternate names, we'll treat it like a toggle button, not a check box.
			altnames = True
		else
			states = " " +sutil__msgCheckBox +states
		endIf
	elif iRole == Role_System_CheckButton
		states = " " +sutil__msgRadioButton +states
	endIf
	if iState & State_System_Checked
		if altnames
			name = stringSegment(name, "/", 2)
		else
			states = states +" " +cMsg_Checked
		endIf
	else
		if altnames
			name = stringSegment(name, "/", 1)
		else
			states = states +" " +cMsg_NotChecked
		endIf
	endIf
elif iRole == Role_System_Separator
endIf
if states
	name = name +" -" +states
endIf
return name
endFunction

string function SkypeWatch__sutil__myGetWindowText(handle hwnd)
; Returns the text in the given window with line breaks and without graphics.
var
	int oldG,
	string buf
oldG = getJCFOption(Opt_Include_Graphics)
setJCFOption(Opt_Include_Graphics, 0)
buf = GetTextInRect (getWindowLeft(hwnd), getWindowTop(hwnd), getWindowRight(hwnd), getWindowBottom(hwnd), 0, IgnoreColor, IgnoreColor, True)
setJCFOption(Opt_Include_Graphics, oldG)
return buf
endFunction

int function SkypeWatch__sutil__versionCompare(string vcode)
; Return an integer indicating how the actual Skype version compares to the given one:
; -1 means Skype is older, 0 means they're equal, and 1 means Skype is newer.
; -2 means the real Skype version code is not known.
; vcode: A string Skype version, like 5.10.
if stringLeft(vcode, 1) != "v"
	vcode = formatString("v%1", vcode)
endIf
var string real = SkypeWatch__sutil__getSkypeVersionCode(False)
if !real
	return -2
elif stringCompare(real, vcode, True) == 0
	return 0
endIf
vcode = stringChopLeft(vcode, 1)
real = stringChopLeft(real, 1)
var int n = max(
	SkypeWatch__string__segmentCount(real, "."),
	SkypeWatch__string__segmentCount(vcode, ".")
)
var int i = 1
var string vseg, string rseg, int vint, int rint
while i <= n
	vseg = stringSegment(vcode, ".", i)
	if !stringLength(vseg)
		; This means versionCompare("5.9") returns equal for Skype 5.9.x.
		return 0
	endIf
	rseg = stringSegment(real, ".", i)
	if !stringLength(rseg)
		; versionCompare("5.9.1.234") returns -1 if Skype's version code only comes back "5.9."
		return -1
	endIf
	vint = stringToInt(vseg)
	rint = stringToInt(rseg)
	if vint > rint
		return -1
	elif vint < rint
		return 1
	; Else equal, so keep looking.
	endIf
	i = i +1
endWhile
; And if the loop ran out, all is equal.
return 0
endFunction

globals
	int SkypeWatch__sutil___vcLastTicks,
	string SkypeWatch__sutil___vcLastVer
string function SkypeWatch__sutil__getSkypeVersionCode(int force)
; Returns a code that indicates what Skype version is running on this machine.
; Codes returned will be exact (if possible) or approximate version numbers of Skype.exe.
; Approximations will always be greater than or equal to the actual version number of skype.exe and will be suffixed with "-" to indicate that they are approximations.
; Returns the null string if the version code can't be obtained or approximated.
; The returned code is prefixed with "v" because this makes the < and > operators behave reasonably enough with version codes.
; WARNING:  "Reasonably enough" means functional but still odd: "v1.3" == "v1.3.2"; == is apparently tried in JAWS before < and >.
; Note: JAWS' GetVersionInfoStrings function is unable to return the real version code from some Skype versions.
; In such a case, the Scripting.FileSystemObject is used to get it.
; Note:  In JAWS versions older than 7.10, the full version code can't be obtained except when Skype is in focus.
; This code is also used by SkypeWatch, so that's not practical in all cases.
; The plan is to get the real code when possible and cache it wisely (I hope).
var
	handle hwnd,
	int ticks, int tickDiff,
	string appPath,
	object o

; Only actually check for the version every 10 seconds, to minimize work.
; This function can be called very frequently (e.g., if used from FindCaret), so this removes the need to worry about that.
ticks = getTickCount()
tickDiff = ticks -SkypeWatch__sutil___vcLastTicks
if !force && tickDiff <= 10000 && stringLength(SkypeWatch__sutil___vcLastVer)
	return SkypeWatch__sutil___vcLastVer
endIf
SkypeWatch__sutil___vcLastTicks = ticks

; If possible, get the application path, which allows us to get the real version number.
; This requires some part of Skype to be in focus.
appPath = ""
appPath = SkypeWatch__string__ref(getWindowOwner(SkypeWatch__appwin__getSkypeWindow(appwin__SWT_Main)))
if !appPath
	appPath = getAppFilePath()
	if !stringContains(stringLower(appPath), "skype.exe")
		appPath = ""
	endIf
endIf

; If we have a path, use it.
if appPath
	; First try JAWS' way, since it seems to work in later Skypes again.
	SkypeWatch__sutil___vcLastVer = getVersionInfoString(appPath, "FileVersion")
	if stringLength(SkypeWatch__sutil___vcLastVer) > 5
		SkypeWatch__sutil___vcLastVer = "v" +SkypeWatch__sutil___vcLastVer
		return SkypeWatch__sutil___vcLastVer
	endIf
	; Get the actual version code from the fileSystemObject if it feels like cooperating.
	o = SkypeWatch__util__getFSObject()
	if o
		SkypeWatch__sutil___vcLastVer = o.getFileVersion(appPath)
		if stringLength(SkypeWatch__sutil___vcLastVer)
			SkypeWatch__sutil___vcLastVer = "v" +SkypeWatch__sutil___vcLastVer
			return SkypeWatch__sutil___vcLastVer
		endIf
	endIf
endIf

; Favor the newest Skype available when we can't tell what it really is.
; Also force an update until we get a better answer though.
SkypeWatch__sutil___vcLastVer = "v999-"
SkypeWatch__sutil___vcLastTicks = 0
return SkypeWatch__sutil___vcLastVer
endFunction

string function SkypeWatch__sutil__getSkypeUserName()
; Returns the Skype ID of the current Skype user.
var
	handle hwnd,
	int pos,
	string buf
hwnd = SkypeWatch__appwin__getSkypeWindow(appwin__SWT_Main)
if !hwnd
	return ""
endIf
buf = getWindowName(hwnd)
pos = stringContainsFromRight(buf, " - ")
if !buf
	return ""
endIf
buf = stringChopLeft(buf, pos+2)
return stringTrimLeadingBlanks(stringTrimTrailingBlanks(buf))
endFunction

string function SkypeWatch__sutil__gotoSkypeScriptPage(string page, int deferred)
; "Runs" or returns the run command for the given page.
; Returns the null string if deferred is False and the command to run if deferred is True.
; page: The page to launch.
; deferred: False to run, True to return the command to be run as a string.
var
	string strURLString,
	string strBrailleDriver, string strBrailleDisplay, string strBraillePort,
	string strSkypeProgramPath,
	string strSkypeProgramDate

if deferred
	var string cmd
	cmd = "SkypeWatch__sutil__gotoSkypeScriptPage(\"" +page +"\", 0)"
	return cmd
endIf

strSkypeProgramPath = getAppFilePath()
strSkypeProgramDate = IntToString(GetFileDate (strSkypeProgramPath))
BrailleGetDeviceInfo(strBrailleDriver, strBrailleDisplay, strBraillePort)

strURLString = sutil___url +page
+"?JFWVersion="+IntToString(getJFWVersion())
+"&ScriptVersion="+sutil__Script_Version
+"&ScriptRevision="+sutil__Script_Revision
+"&BrailleDisplay="+strBrailleDisplay
+"&SkypeProgramDate="+strSkypeProgramDate
+"&SkypeProgramVersion=" +stringChopLeft(SkypeWatch__sutil__getSkypeVersionCode(True), 1)
+"&WindowsOSVersion="+IntToString(GetWindowsOS())
+"&SkypeWatch="+intToString(SkypeWatch__sutil__isSkypeWatchLoaded())
+"&SkypeUserName="+SkypeWatch__sutil__getSkypeUserName()
strURLString = stringReplaceSubstrings(strURLString, " ", "%20")

if UserBufferIsActive () UserBufferDeactivate() EndIf
SayFormattedMessage (OT_Message, sutil__msgSkypeScriptsHomePage_L, sutil__msgSkypeScriptsHomePage_S)
Run("\"" +strURLString +"\"")
endFunction

int function SkypeWatch__sutil___splitContactInfo(string buffer, string byRef name, string byRef id, string byRef stat)
; Returns by reference the contact name, Skype ID, and status from the given buffer.
; The buffer should have come directly from MSAA's AccHelp property.
; Buffer format: <contactName>[ (<SkypeID>)] - <statusInfo>
; Returns True on success and False on failure.
; Author:  Doug Lee
var
	int pos

if stringIsBlank(buffer)
	; Happens on contact list entries for multichats.
	beep()
	return False
endIf

; Start by removing and collecting the status text (Online, Offline and last-seen time, etc.).
pos = stringContainsFromRight(buffer, "/>")  ; should always be found
if !pos
	return False
endIf
stat = stringChopLeft(buffer, pos+3)
buffer = stringLeft(buffer, stringContainsFromRight(buffer, "<")-2)
; Current buffer format: <contactName>[ (<SkypeID>)].
buffer = stringSegment(buffer, "\13", 1)

; Find out if there's a second name, and set name and id accordingly.
; If there's not, all we have is the Skype ID, which can't contain parentheses.
; If the display name differs from the Skype ID, we'll have both here.
pos = stringContainsFromRight(buffer, "(")  ; only found if second name is present
if pos
	name = stringLeft(buffer, pos-2)
	id = substring(buffer, pos+1, stringLength(buffer)-pos-1)
else
	name = buffer
	id = buffer
endIf
return True
endFunction

int function SkypeWatch__sutil___isPhoneNumber(string s)
; Returns True if the given string appears to be a phone number as might be found in Skype.
var
	int i, int n, string ch
s = stringTrimLeadingBlanks(stringTrimTrailingBlanks(s))
n = stringLength(s)
if n > 50
	; 50 is a pretty liberal limit for the length of a phone number...
	return False
endIf
i = 1
if stringLeft(s, 1) == "+"
	i = 2
endIf
while i <= n
	ch = substring(s, i, 1)
	if !stringIsBlank(ch) && !stringContains(",-.0123456789", ch)
		return False
	endIf
	i = i +1
endWhile
return True
endFunction

string function SkypeWatch__sutil__removeHTML(string s)
; Remove HTML tags and (some) character names from s.
var
	string lhs, string rhs,
	int pos,
	string result
pos = stringContains(s, "<")
while pos
	result = result +" " +stringLeft(s, pos-1)
	s = stringChopLeft(s, pos)
	pos = stringContains(s, ">")
	if pos
		s = stringChopLeft(s, pos)
	else
		s = ""
	endIf
	pos = stringContains(s, "<")
endWhile
result = stringChopLeft(result +" " +s, 1)
SkypeWatch__string__replace(result, "&nbsp;", " ", "e")
SkypeWatch__string__replace(result, "&lt;", "<", "e")
SkypeWatch__string__replace(result, "&gt;", ">", "e")
SkypeWatch__string__replace(result, "&lquot;", "\"", "e")
SkypeWatch__string__replace(result, "&rquot;", "\"", "e")
return result
endFunction

collection function SkypeWatch__sutil__getSkypeContactInfoFromMSAA(object o, int childID)
; Get as much information from MSAA about the indicated contact as possible.
; Currently this includes name, ID, status, ExtStatus, SpeedDial, and MoodText.
var
	string buf, string name, string ID, string stat, string extstat, string mood,
	int pos,
	collection cContact
name = ""
id = ""
stat = ""
extstat = ""

buf = SkypeWatch__sutil__removeHTML(o.accHelp(childID))
if stringLength(buf) && SkypeWatch__sutil___splitContactInfo(buf, name, id, stat)
	; Separate "Last seen: ..." from "Offline" when found.
	pos = stringContainsFromRight(stat, "(")
	if pos
		extstat = stringChopLeft(stat, pos)
		extstat = stringChopRight(extstat, 1)  ; chops off ')'
		stat = stringTrimTrailingBlanks(stringLeft(stat, pos-1))
	endIf
else
	; Use accValue when there's no accHelp.
	; AccHelp for contacts vanished around Skype 4.0 beta 3.
	buf = o.accValue(childID)
	if stringLength(buf) && stringRight(buf, 1) == ")"
		pos = stringContainsFromRight(buf, "(")
		if pos
			name = stringTrimTrailingBlanks(stringLeft(buf, pos-1))
			id = stringChopRight(stringChopLeft(buf, pos), 1)
		endIf
	endIf
endIf

cContact = new collection
cContact.Name = name
cContact.ID = id
cContact.Status = stat
if extstat
	cContact.ExtStat = extstat
endIf
; Include extra stuff from accDescription if found (Speed dial for example).
; Known format as of Skype 2.0.0.78: <status>[, <moodText>][ Speed dial <speedDialNumber>]
; TODO:  See if "Speed dial" is localized in accDescription.
buf = stringTrimTrailingBlanks(o.accDescription(childID))
if buf
	if buf == stat || !stringLength(stat)   ; match up to the end of status text
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(stat)))
	else
		; accDescription should always begin with status text.
		beep()
	endIf
	mood = ""
	if buf == ", "
		; There is mood text.
		buf = stringChopLeft(buf, 2)
		pos = stringContainsFromRight(buf, sutil__MSAASpeedDial)
		if pos
			mood = stringLeft(buf, pos-1)
			buf = stringChopLeft(buf, pos)
		else
			mood = buf +""
			buf = ""
		endIf
	endIf
	if mood
		cContact.MoodText = mood
	endIf
	if buf == sutil__MSAASpeedDial
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(sutil__MSAASpeedDial)))
		cContact.SpeedDial = intToString(stringToInt(buf))
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(intToString(stringToInt(buf)))))
	endIf
	if !stringIsBlank(buf)
		cContact.Other = buf
	endIf
endIf  ; non-empty buf from accDescription
return cContact
endFunction

int function SkypeWatch__sutil___refreshSkypeWindow(handle hwnd)
if False && getWindowClass(getFirstChild(hwnd)) == "TScrollBar"
	;speechOff()
	setFocus(getFirstChild(hwnd))
	pause()
	setFocus(hwnd)
	pause()
	speechOn()
	return True
endIf
;sayInteger(sendMessage(hwnd, 14, 0, 0))
return False
endFunction

collection function SkypeWatch__sutil__getCurrentContactInfo()
; Returns a collection consisting of all retrievable fields about the current contact, with field names.  Both MSAA and screen text are used.
; When a field appears on screen but its name can't be determined, it is named "Field" plus its ordinal number in the field list (e.g., "Field5").
; Currently recognized fields are Name, ID, Status, ExtStatus, SpeedDial, MoodText, Location, Time, TimeZone, and Phone1-3.
var
	handle hContactList,
	object o, int childID,
	int tries,
	int left, int right, int top, int bottom,
	int oldRestriction,
	int i, int ibase, int phoneCount, int done,
	string buf, int pos,
	string mood, string stat, string id,
	collection cContact

; First get the MSAA information.
o = SkypeWatch__sutil__getFocusObject(childID)
if !o
	o = SkypeWatch__appwin__getMSAAObject(appwin__SWT_ContactList)
endIf
if !o
	return ""
endIf
childID = o.accFocus
if !childID
	return ""
endIf
cContact = SkypeWatch__sutil__getSkypeContactInfoFromMSAA(o, childID)

; Now restrict the invisible cursor to the contact's rectangle as indicated by its MSAA coordinates.
; Obscured windows are skipped because that means the information from the screen might not be valid anyway.
; This whole step is also skipped if focus is not in the contact list, because even other contact lists with the same MSAA structure do not share the same on-screen format.
; Example:  the participant list in a chat window (Skype 3-).
hContactList = SkypeWatch__appwin__getSkypeWindow(appwin__SWT_ContactList)
if !hContactList
	; TODO: Exits here for Skype 5 because there is no contact list window.
	return cContact
endIf
saveCursor() invisibleCursor() saveCursor()
oldRestriction = getRestriction()
tries = 2
done = False
while !done
	if !tries || getFocus() != hContactList || isWindowObscured(hContactList) || !SkypeWatch__sutil__MSAA__rect(o, childID, left, right, top, bottom)
		;beep()
		;sayString("Rect " +getTextInRect(left, top, right, bottom, 0, IgnoreColor, IgnoreColor, True))
		setRestriction(oldRestriction)
		return cContact
	endIf
	moveTo((left+right)/2, (top+bottom)/2)
	setRestrictionToRect(left, top, right, bottom)
	; Make sure we're on the first text chunk in the rectangle.
	JAWSPageUp() JAWSHome() nextChunk() priorChunk() priorChunk()
	if stringContains(getLine(), cContact.Name)
		done = True  ; causes the loop to exit
	else
		if tries == 2
			SkypeWatch__sutil___refreshSkypeWindow(hContactList)
		endIf
		pause()
		;sayString("retry")
	endIf
	tries = tries -1
endWhile

; Collect the data one text chunk at a time from the screen.
; The first line is skipped in all cases because it is just the contact name, which we already got from MSAA.
; For SkypeOut contacts, the phone number appears on the second line, but we already have that from MSAA also.
; Field names can be filled in retroactively by this code as context is discovered.
if nextLine()
	JAWSHome()
	stat = cContact.Status
	id = cContact.ID
	if stringContains(stringLower(stat), sutil__MSAASkypeOut) && getChunk() == ID
		if !nextChunk()
			setRestriction(oldRestriction)
			return cContact
		endIf
	endIf
	done = False
	ibase = collectionItemCount(cContact)
	i = ibase
	mood = cContact.MoodText
	while !done && i < 50
		buf = getChunk()
		if stringCompare(buf, sutil__scMenu, True) && stringCompare(buf, mood, True)
		pos = stringContainsFromRight(buf, "("+sutil__scGMT)
		if pos
			; Skype 3.x structure: <time> (<tz>)[, <location>]
			ccontact.Time = stringTrimTrailingBlanks(stringLeft(buf, pos-1))
			buf = stringChopLeft(buf, pos)  ; also chops off '('
			pos = stringContains(buf, ")")
			cContact.TimeZone = stringLeft(buf, pos-1)
			buf = stringChopLeft(buf, pos)
			if buf == ","
				cContact.Location = stringTrimLeadingBlanks(stringChopLeft(buf, 1))
			endIf
		else  ; not a time/location line
			if SkypeWatch__sutil___isPhoneNumber(buf)
				phoneCount = phoneCount +1
				; This arranges for a single phone number to be called "Phone" but multiple phone numbers to be called "Phone1," "Phone2," etc.
				if phoneCount == 1
					cContact.Phone = buf
				elif phoneCount == 2
					cContact["Phone"+intToString(phoneCount)] = buf
					cContact.Phone1 = cContact.Phone
					collectionItemRemove(cContact, "Phone")
				else
					cContact["Phone"+intToString(phoneCount)] = buf
				endIf
			else
				; We don't know the name of this field, so we'll assign a generic one (and hope a better one is found above soon).
				cContact["Field"+intToString(i/2+1)] = buf
			endIf
			i = i +2
		endIf
		endIf  ; chunk is "Menu"
		done = !nextChunk()
	endWhile
endIf
setRestriction(oldRestriction)
return cContact
endFunction

globals string SkypeWatch__sutil___lastStat
string function SkypeWatch__sutil__getMyOnlineStatus()
; Returns this user's online/offline/etc. status as a string.
; In Skype 3.2+, the return string is already localized, as it comes directly from MSAA.
; Author:  Doug Lee
var
	string sVer,
	object o, int childID, int n,
	int found,
	string stat
o = SkypeWatch__appwin__getMSAAObject(appwin__SWT_Myself)
if o
	n = o.accChildCount
	childID = 1
	found = False
	while childID <= n
		if o.accRole(childID) == Role_System_ButtonDropDown
			stat = o.accName(childID)
			if !stat
				stat = sutil__msgStatUnknown
			endIf
			SkypeWatch__sutil___lastStat = stringLower(stat)
			found = True
		endIf
		childID = childID +1
	endWhile
endIf
return SkypeWatch__sutil___lastStat
endFunction

string function SkypeWatch__sutil__speakableSkypeID(string id)
; Speakable version of the given Skype ID:
;	- Conference IDs say "conference."
;	- US phone numbers are formatted as +1 nnn-nnn-nnnn.
if id == "#"
	id = sutil__msgConference
elif id == "+" && stringLength(id) == 12
	id = formatString("%1 %2-%3-%4",
		stringLeft(id, 2),
		substring(id, 3, 3),
		substring(id, 6, 3),
		stringRight(id, 4)
	)
endIf
return id
endFunction

int function SkypeWatch__sutil__play(string sound)
; Play the indicated sound file.
var
	string spath
;spath = findJAWSSoundFile(sound)
spath = getJAWSSettingsDirectory() +"\\sounds\\" +sound
if !fileExists(spath)
	spath = getJAWSSettingsDirectory() +"\\" +sound
endIf
if !fileExists(spath)
	spath = findJAWSSoundFile(sound)
endIf
if spath
	playSound(spath)
endIf
endFunction

string function SkypeWatch__sutil__getWindowClassPath(handle hwnd)
var
	string path,
	string sep
sep = "/"
while hwnd
	path = getWindowClass(hwnd) +"/" +path
	hwnd = getParent(hwnd)
endWhile
return stringChopRight(path, 1)
endFunction

int function SkypeWatch__sutil__isUnobscured(object o, int childID)
; Returns True if the indicated MSAA element appears to be unobscured by something else.
; Uses GetObjectAtPoint, so requires .AccHitTest to work.
var
	int left, int right, int top, int bottom,
	object o1, int childID1
if !SkypeWatch__sutil__MSAA__rect(o, childID, left, right, top, bottom)
	; If it doesn't have screen space, it's sort of obscured by default.
	return False
endIf
o1 = getObjectAtPoint(childID1, (left+right)/2, (top+bottom)/2)
if childID1 == childID && o.accName(0) == o1.accName(0)
	return True
endIf
return False
endFunction

int function SkypeWatch__sutil__isAncestorOf(handle hwndChild, handle hwndAncestor)
; Returns True if and only if hwndChild descends at some level from hwndAncestor.
; True is also returned if the two are equal.
while hwndChild
	if hwndChild == hwndAncestor
		return True
	endIf
	hwndChild = getParent(hwndChild)
endWhile
return False
endFunction

int function SkypeWatch__sutil__getMSAAItemDim(object o, int childID, string which)
; Return left, right, top, bottom, width, or height for the given MSAA item.
var
	int left, int top, int width, int height
o.accLocation(intRef(left), intRef(top), intRef(width), intRef(height), childId)
if !left && !top && !width && !height
	return 0
endIf
left = left +1
top = top +1
if which == "l" return left
elif which == "t" return top
elif which == "w" return width
elif which == "h" return height
elif which == "r" return left+width
elif which == "b" return top+height
else return 0
endIf
endFunction

int function SkypeWatch__sutil__findNthColorTransition(handle hwnd, int delta, int n)
; Finds the nth color transition in delta's direction from the current active cursor's location.
; Returns True on success and False on failure.
; The active cursor is moved to the found spot only if a spot is found.
; The search is limited by the edge of hwnd in delta's direction if hwnd is not null.
; Author:  Doug Lee
var
	int x, int y,
	int limit, int tooFar,
	int clr, int clr0
if hwnd
	if delta > 0
		limit = getWindowRight(hwnd)
	else
		limit = getWindowLeft(hwnd)
	endIf
else
	if delta > 0
		limit = screenGetWidth() +1
	else
		limit = 0
	endIf
endIf
x = getCursorCol() +(delta *3)  ; arbitrary safety margin to avoid edge of text characters
y = getCursorRow()
clr0 = getColorAtPoint(x, y)
tooFar = False
while n && !tooFar
	clr = getColorAtPoint(x, y)
	if clr != clr0
		n = n -1
		clr0 = clr
	endIf
	x = x +delta
	if delta > 1
		tooFar = (x > limit)
	else
		tooFar = (x < limit)
	endIf
endWhile
if tooFar
	return False
endIf
moveTo(x, y)
return True
endFunction


;------------------------------- Module winfinder ------------------------------
; Ways to search for windows by typeCode, class substring match, and by calling test functions.
;
; Author:  Doug Lee

globals
; Variables used to communicate between SkypeWatch__sutil__winfinder__findWindow and SkypeWatch__sutil__winfinder___findWindowHelper.
; Required in order to use enumerateChildWindows().
	handle SkypeWatch__sutil__winfinder___windowFound,
	string SkypeWatch__sutil__winfinder___windowClass,
	string SkypeWatch__sutil__winfinder___windowName

int function SkypeWatch__sutil__winfinder___findWindowHelper(handle hwnd)
; Called via EnumerateChildWindows.
var
	int pos,
	string wclass,
	int found
wclass = SkypeWatch__sutil__winfinder___windowClass
if stringContains("@!", stringLeft(wclass, 1))
	pos = stringContains(wclass, "%%")
	if pos
		wclass = stringLeft(wclass, pos-1) +intToString(hwnd) +stringChopLeft(wclass, pos+1)
	endIf
endIf
if wclass == "@"
	found = stringToInt(formatStringWithEmbeddedFunctions("<" +stringChopLeft(wclass, 1) +">"))
elif wclass == "!"
	found = !stringToInt(formatStringWithEmbeddedFunctions("<" +stringChopLeft(wclass, 1) +">"))
else
	found = (!stringCompare(getWindowClass(hwnd), wclass, False) && !stringCompare(getWindowName(hwnd), SkypeWatch__sutil__winfinder___windowName, False))
endIf
if found
	SkypeWatch__sutil__winfinder___windowFound = hwnd
	return False
endIf
return True
endFunction

handle function SkypeWatch__sutil__winfinder__findWindow(handle hTop, string wclass, string wname)
; Generalized window finder. Can be used just like JAWS' FindWindow function.
; Also accepts in wclass @funcCall or !funcCall to use arbitrary function calls to test windows.
; funcCall should be a function call with %% in place of the hwnd parameter.
; NOTE: The Name option uses getWindowName, not WM_GETTEXT. They can differ.
; The JAWS built-in FindWindow function uses WM_GETTEXT according to impirical data.
var
handle hwnd,
	handle nullHandle
hwnd = findWindow(hTop, wclass, wname)
if hwnd
	return hwnd
endIf
SkypeWatch__sutil__winfinder___windowClass = wclass
SkypeWatch__sutil__winfinder___windowName = wname
SkypeWatch__sutil__winfinder___windowFound = nullHandle
EnumerateChildWindows (hTop, "SkypeWatch__sutil__winfinder___findWindowHelper")
return SkypeWatch__sutil__winfinder___windowFound
endFunction

int function SkypeWatch__sutil__winfinder___isTypeCode(handle hwnd, int typeCode, int isSubtype)
; Returns True if the given window has the given type or subtype code.
if isSubtype
	return (getWindowSubtypeCode(hwnd) == typeCode)
else
	return (getWindowTypeCode(hwnd) == typeCode)
endIf
endFunction

handle function SkypeWatch__sutil__winfinder__findWindowByTypeCode(handle hTop, int typeCode, int isSubtype)
; Finds the first window under hTop (inclusive) having the given type or subtype code.
; Author:  Doug Lee
return findWindowByType(hTop, typeCode)
endFunction

int function SkypeWatch__sutil__winfinder__getWindowRole(handle hwnd)
; Returns the given window's MSAA client object's role.
; This role, if not Role_System_Client, should represent the window's actual function.
; Author:  Doug Lee
var int childID  ; required by GetObjectFromEvent but otherwise ignored here
return getObjectFromEvent(hwnd, -4, 0, childID).accRole(0)
endFunction

int function SkypeWatch__sutil__winfinder__getWindowValue(handle hwnd)
; Returns the given window's MSAA client object's value.
; Author:  Doug Lee
var int childID  ; required by GetObjectFromEvent but otherwise ignored here
return getObjectFromEvent(hwnd, -4, 0, childID).accValue(0)
endFunction

int function SkypeWatch__sutil__winfinder___isMSAARole(handle hwnd, int iRole)
; Returns True if the given window's MSAA client object has the given role.
; Author:  Doug Lee
var int childID
return (SkypeWatch__sutil__winfinder__getWindowRole(hwnd) == iRole)
endFunction

handle function SkypeWatch__sutil__winfinder__findWindowByMSAARole(handle hTop, int iRole)
; Finds the first window under hTop (inclusive) whose MSAA client object has the given role.
; Author:  Doug Lee
return SkypeWatch__sutil__winfinder__findWindow(hTop, "@SkypeWatch__sutil__winfinder___isMSAARole(%%, " +intToString(iRole) +")", "")
endFunction

int function SkypeWatch__sutil__winfinder___isClassMatch(handle hwnd, string classPart)
; Returns True if the given window's class name contains classPart.
if stringContains(stringLower(getWindowClass(hwnd)), stringLower(classPart))
	return True
endIf
return False
endFunction

handle function SkypeWatch__sutil__winfinder__findWindowByClassSubstring(handle hTop, string classPart)
; Finds the first window under hTop (inclusive) whose class name contains classPart.
; Author:  Doug Lee
return SkypeWatch__sutil__winfinder__findWindow(hTop, "@SkypeWatch__sutil__winfinder___isClassMatch(%%, " +classPart +")", "")
endFunction

;----------------------------- End module winfinder ----------------------------


;--------------------------------- Module MSAA ---------------------------------
; Miscellaneous functions for MSAA access
;
; Author:  Doug Lee

collection function SkypeWatch__sutil__MSAA__rect(object o, int childID, optional int byRef left, int byRef right, int byRef top, int byRef bottom)
; Get the bounding rectangle for the given MSAA object and childID.
; The rectangle is returned 1-based, as JAWS likes its rectangles.
; MSAA provides 0-based left/top/width/height.
; The rectangle is returned by reference and also as a collection. Returns null on failure.
if !o
	left = 0
	right = 0
	top = 0
	bottom = 0
	return
endIf
if !childID
	childID = 0
endIf
o.accLocation(intRef(left), intRef(top), intRef(right), intRef(bottom), childID)
if !left && !right && !top && !bottom
	return
endIf
; but we want 1-based left/right/top/bottom.
left = left +1
top = top +1
right = left +right
bottom = top +bottom
var collection c c = new collection
c.left = left
c.right = right
c.top = top
c.bottom = bottom
return c
endFunction

;------------------------------- End module MSAA -------------------------------

;====================== End module sutil (from SkypeUtil) ======================


;================================== Module opt =================================
; User option code.

const
; A constant used to detect missing, as opposed to set-to-empty, .ini keys.
	opt___NoVal = 12345678

variant function SkypeWatch__opt__getUserOption(string varName)
; Retrieves and returns the value of the indicated user-settable variable from skype.ini.
; This function makes sure to return integer 0 for the integer value 0, rather than returning the string "0" which shows up as True in a boolean test.
var
	string val
; Chop off descriptions.
varname = stringSegment(varname, ":", 1)
val = iniReadString(sutil__INISectName, varname, intToString(opt___NoVal), sutil__INIFileName)
; Favor integer 0 (False) rather than string "0" (True).
if stringContains(stringLeft(stringTrimLeadingBlanks(val), 1), "0")
	return ""
endIf
if val == opt___NoVal
	; All default values for options go here.
	if varname == opt__SV_RecognizeChat
		val = "1"
	elif varname == opt__SV_ChatMessageTimes
		val = "1"
	elif varname == opt__SV_SpeakOnlineStatus
		val = "1"
	elif varname == opt__SV_TypingAnnounce
		val = "1"
	elif varname == opt__SV_TypingSounder
		val = "1"
	elif varname == opt__SV_SkypeCommandKey
		val = "["
	elif varname == opt__SV_SkypeWatchMessages
		val = "1"
	else
		val = ""
	endIf
endIf
return val
endFunction

int function SkypeWatch__opt__setSkypeVar(string varName, variant val)
; Sets the indicated user-settable variable in skype.ini to the given value.
; Chop off descriptions.
varname = stringSegment(varname, ":", 1)
return iniWriteString(sutil__INISectName, varname, val, sutil__INIFileName)
endFunction

variant function SkypeWatch__opt__userOptionCycler(string varname, string choices, int iRetCurVal)
; Used by DlgSelectFunctionToRun cycler/toggler functions to cycle/toggle variable values.
; Varname should be the option name being handled (SV_* constant).
; Choices should be a list of text values, separated by vertical bars (|).
; iRetCurVal should be the integer taken by the calling toggler/cycler function.
; The return value of this function behaves according to specifications for DlgSelectFunctionToRun functions and should be returned directly by the toggler/cycler function.
; The integer value saved in skype.ini will be the 0-based index of the chosen option.
var
	int i,
	int n
i = SkypeWatch__opt__getUserOption(varname)
n = SkypeWatch__string__segmentCount(choices, "|")
if !iRetCurVal
	; Update it.
	i = i +1
	if i >= n
		i = 0
	endIf
	SkypeWatch__opt__setSkypeVar(varname, i)
endIf
; i+1 rather than just i because i is 0-based and stringSegment is 1-based.
return stringSegment(choices, "|", i+1)
endFunction

;================================ End module opt ===============================


;================================ Module appwin ================================
; getSkypeWindow and related app window identification code.

globals
; The main Skype window
	handle SkypeWatch__appwin___mainWindow,
; A place to cache the MSAA contact list object.
	object SkypeWatch__appwin___contactListObject

int function SkypeWatch__appwin__isMainSkypeWindow(handle hwnd)
var
	string wclass
wclass = getWindowClass(hwnd)
if wclass == "TskMainForm"
|| wclass == "TZapMainForm"
	return True
endIf
return False
endFunction

handle function SkypeWatch__appwin__getMainSkypeWindow()
; Returns the top-level main Skype window if possible.
; If the main window cannot be found, also clears the MSAA contact list object cache.
var
	object nullObject,
	handle hwnd
hwnd = getTopLevelWindow(getFocus())
if SkypeWatch__appwin__isMainSkypeWindow(hwnd)
	SkypeWatch__appwin___mainWindow = hwnd
	return hwnd
endIf
hwnd = SkypeWatch__appwin___mainWindow
if SkypeWatch__appwin__isMainSkypeWindow(hwnd)
	return hwnd
endIf
hwnd = findTopLevelWindow("tSkMainForm", "")
if !hwnd
	hwnd = findTopLevelWindow("tSkMainForm.UnicodeClass", "")
endIf
if hwnd
	SkypeWatch__appwin___mainWindow = hwnd
	return hwnd
endIf
SkypeWatch__appwin___contactListObject = nullObject
return 0
endFunction

int function SkypeWatch__appwin__isChatWindow(handle hwnd)
; Returns 1 (True) if the given window is or is within a Skype Instant Messaging window.
while hwnd
	if getWindowClass(hwnd) == "TConversationForm"
		return 1
	endIf
	hwnd = getParent(hwnd)
endWhile
return 0
endFunction

int function SkypeWatch__appwin___isSkypeToolTip(handle hwnd)
; Returns True if the given window is the Skype tooltip window.
return hwnd && stringContains(stringLower(getWindowClass(hwnd)), appwin___cwsToolTip)
		;&& getParent(hwnd)
		;&& getWindowClass(getParent(hwnd)) == appwin___wcApp && getWindowName(getParent(hwnd)) == appwin__wnSkype
endFunction

int function SkypeWatch__appwin__isSkypeNotifyWindow(handle hwnd)
; Returns True if the given window is a Skype system tray notification window or call alert.
var
	string wclass
wclass = getWindowClass(hwnd)
return (wclass == "TCallNotificationWindow"
|| wclass == "TTrayAlert")
endFunction

handle function SkypeWatch__appwin__findFirstSkypeNotifyWindow()
; Find the first available Skype notification window and return its handle.
; Returns null on failure.
var handle h1 = findTopLevelWindow("TTrayAlert", "")
var handle h2 = findTopLevelWindow("TCallNotificationWindow", "")
if h1 && h2
&& getWindowHierarchyX(h2) < getWindowHierarchyX(h1)
	h1 = h2
elif !h1
	h1 = h2
endIf
return h1
endFunction

int function SkypeWatch__appwin___isRequestedSkypeWindow(handle hwnd, string which)
; Returns True if the given window is the given Skype window.
if which == appwin__SWT_ToolTip
	return SkypeWatch__appwin___isSkypeToolTip(hwnd)
elif which == appwin__SWT_Notify
	return SkypeWatch__appwin__isSkypeNotifyWindow(hwnd)
else
	return False
endIf
endFunction

int function SkypeWatch__appwin__isSkypeWindow(handle hwnd)
; Returns True if the given window is part of Skype.
var
	string wclass
if stringRight(SkypeWatch__string__ref(getWindowOwner(hwnd)), 9) == "Skype.exe"
	return True
endIf
return False
endFunction

handle function SkypeWatch__appwin___findTopLevelSkypeWindow(string which)
; Returns the handle of the requested Skype top-level window.
; This window may not exist, such as right after Skype loads.
var
	handle hwnd
hwnd = SkypeWatch__util__getFirstTopLevelWindow()
while hwnd && !SkypeWatch__appwin___isRequestedSkypeWindow(hwnd, which)
	hwnd = getNextWindow(hwnd)
endWhile
if SkypeWatch__appwin___isRequestedSkypeWindow(hwnd, which)
	return hwnd
endIf
return 0
endFunction

handle function SkypeWatch__appwin__getActiveChatWindow(handle hTop)
var
	handle hwnd
hwnd = findWindow(hTop, "TConversationForm", "")
if !hwnd
	hwnd = findWindow(hTop, "TConversationForm.UnicodeClass", "")
endIf
if !hwnd || isWindowVisible(hwnd)
	; Got it or there isn't such a thing.
	return hwnd
endIf
hwnd = getNextWindow(hwnd)
while hwnd
	if isWindowVisible(hwnd) && SkypeWatch__appwin__isChatWindow(hwnd)
		return hwnd
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

handle function SkypeWatch__appwin__getSkypeWindow(string which)
; Returns the handle of the window of the indicated type.
; Note that this only works for window types that are unique.
; This function should be kept in sync with the set of SWT_* constants in skype.jsh.
var
	handle hMain, handle hwnd, handle hwnd1,
	string hList
if which == "?"
	hList = appwin__SWT_Main +"|" +appwin__SWT_Myself
		+"|" +appwin__SWT_ContactList +"|" +appwin__SWT_ConversationList +"|" +appwin__SWT_LiveConversations
		+"|" +appwin__SWT_History
	hList = hList +"|" +appwin__SWT_ChatMain +"|" +appwin__SWT_ChatInput +"|" +appwin__SWT_ChatOutput +"|" +appwin__SWT_ChatList +"|" +appwin__SWT_ChatContact +"|" +appwin__SWT_Communicator +"|" +appwin__SWT_CallQuality
	; Omitted knowingly:  appwin__SWT_ToolTip, appwin__SWT_Notify.
	return hList
endIf
hMain = SkypeWatch__appwin__getMainSkypeWindow()
if !hMain
	return 0
endIf
if which == appwin__SWT_Main
	return hMain
endIf
if which == appwin__SWT_Myself
	return SkypeWatch__sutil__winfinder__findWindowByClassSubstring(hMain, appwin___cwsMyself)
elif which == appwin__SWT_ContactList
	hwnd1 = findWindow(hMain, "TConversationList", "")
	if hwnd1
		if which == appwin__SWT_ContactList
			return hwnd1
		endIf
		return getParent(hwnd1)
	endIf
	return 0
elif which == appwin__SWT_ConversationList
	return findWindow(hMain, "TInboxList", "")
elif which == appwin__SWT_LiveConversations
	hwnd = findWindow(hMain, "TLiveConversationList", "")
	return hwnd
elif which == appwin__SWT_History
	return SkypeWatch__sutil__winfinder__findWindowByClassSubstring(hMain, "history")
elif which == appwin__SWT_ToolTip
	return SkypeWatch__appwin___findTopLevelSkypeWindow(appwin__SWT_ToolTip)
elif which == appwin__SWT_Notify
	var handle hwndn1, handle hwndn2
	hwndn1 = findTopLevelWindow("TCallNotificationWindow", "")
	hwndn2 = findTopLevelWindow("TTrayAlert", "")
	if hwndn1 && hwndn2
		if getWindowHierarchyX(hwndn1) < getWindowHierarchyX(hwndn2)
			hwnd1 = hwndn1
		else
			hwnd1 = hwndn2
		endIf
	elif hwndn1
		hwnd1 = hwndn1
	else
		hwnd1 = hwndn2
	endIf
	return hwnd1
; TZapCommunicator is in each conversation in Compact view,
; but in Default view, it's at the same level as the conversation windows.
elif which == appwin__SWT_Communicator
	hwnd1 = findWindow(hMain, "TZapCommunicator", "")
	if hwnd1
		return hwnd1
	endIf
endIf
if which == appwin__SWT_ChatPrefix || which == appwin__SWT_CallQuality
	; Conversations can be part of the Skype window tree or their own independent tree.
	hMain = getTopLevelWindow(getFocus())
	if !SkypeWatch__appwin__isChatWindow(hMain)
		hMain = SkypeWatch__appwin__getActiveChatWindow(hMain)
	endIf
	if !hMain
		return 0
	endIf
	if which == appwin__SWT_ChatMain
		return hMain
	endIf
	if which == appwin__SWT_ChatInput
		return SkypeWatch__sutil__winfinder__findWindowByClassSubstring(hMain, "TChatRichEdit")
	elif which == appwin__SWT_ChatList
		return findWindow(hMain, "TParticipantList", "")
	elif which == appwin__SWT_ChatContact
		hwnd = findWindow(hMain, "TConversationHeader", "")
		if !hwnd
			hwnd = findWindow(hMain, "TContactProfile", "")
		endIf
		if !hwnd
			hwnd = findWindow(hMain, "TConversationProfile", "")
		endIf
		if !hwnd
			hwnd = findWindow(hMain, "TParticipantList", "")
		endIf
		return hwnd
	elif which == appwin__SWT_ChatOutput
		return findWindow(hMain, "TChatContentControl", "")
	elif which == appwin__SWT_CallQuality
		return findWindow(hMain, "TCallQualityControlHost", "")
	endIf
	return 0
endIf
return 0
endFunction

object function SkypeWatch__appwin__getMSAAObject(string which)
; Returns the requested MSAA object, which should be an SWT_* constant.
; appwin__SWT_MenuBar is special:  It returns the Skype menu bar object but has no corresponding purpose in SkypeWatch__appwin__getSkypeWindow().
var
	object o, int childID, int n,
	handle hwnd
if SkypeWatch__appwin__conv5__applies()
	; In Skype 5.0 Gold, these have MSAA but no window.
	if which == appwin__SWT_ContactList
		return SkypeWatch__appwin__conv5__listByName(appwin__conv5__ListContacts)
	elif which == appwin__SWT_LiveConversations
		return SkypeWatch__appwin__conv5__listByName(appwin__conv5__ListLive)
	elif which == appwin__SWT_ConversationList
		return SkypeWatch__appwin__conv5__listByName(appwin__conv5__ListConversations)
	elif which == appwin__SWT_History
		return SkypeWatch__appwin__conv5__listByName(appwin__conv5__ListHistory)
	elif which == appwin__SWT_ChatOutput
		return SkypeWatch__appwin__conv5__chatContent(SkypeWatch__appwin__getSkypeWindow(appwin__SWT_ChatOutput))
	endIf
endIf
if which != appwin__SWT_MenuBar
	hwnd = SkypeWatch__appwin__getSkypeWindow(which)
	if !hwnd
		return 0
	endIf
	o = getObjectFromEvent(hwnd, -4, 0, childID)  ; Gets the Role_System_Client (or more specific role) object for this window
	return o
endIf
hwnd = SkypeWatch__appwin__getMainSkypeWindow()
if !hwnd
	return 0
endIf
o = getObjectFromEvent(hwnd, 0, 0, childID)  ; Gets the Role_System_Window object for this window
if !o
	return 0
endIf
return o.accChild(-3)  ; application menu bar
endFunction

object function SkypeWatch__appwin__getContactListObject()
; Returns the MSAA object for the contact list, if it can be found.
var
	handle hwnd,
	object o, int childID,
	object nullObject

; Clear out the cached object if Skype is not loaded (SkypeWatch__appwin__getMainSkypeWindow does this).
if !SkypeWatch__appwin__getMainSkypeWindow()
	return nullObject
endIf

; Return the cached object if it is valid.
if SkypeWatch__appwin___contactListObject
	if SkypeWatch__appwin___contactListObject.accRole(0) == Role_System_List
	|| SkypeWatch__appwin___contactListObject.accRole(0) == Role_System_Pane
		return SkypeWatch__appwin___contactListObject
	elif SkypeWatch__appwin___contactListObject.accRole(0) == Role_System_Client
		return nullObject
	endIf
endIf

SkypeWatch__appwin___contactListObject =
SkypeWatch__appwin__conv5__listByName(appwin__conv5__ListContacts)
if SkypeWatch__appwin___contactListObject
	return SkypeWatch__appwin___contactListObject
endIf

hwnd = SkypeWatch__appwin__getSkypeWindow(appwin__SWT_ContactList)
if hwnd
	o = getObjectFromEvent(hwnd, 0-4, 0, childID)
	if !o
		SkypeWatch__appwin___contactListObject = nullObject
		return nullObject
	elif o.accRole(0) != Role_System_List && o.accRole(0) != Role_System_Pane
		SkypeWatch__appwin___contactListObject = nullObject
		return nullObject
	endIf
	SkypeWatch__appwin___contactListObject = o
	return SkypeWatch__appwin___contactListObject
endIf
SkypeWatch__appwin___contactListObject = nullObject
return nullObject
endFunction


;--------------------------------- Module conv5 --------------------------------
; Utilities for managing the Skype 5 beta 2 and Gold, and (hopefully) later conversations control.
; Also includes a function to obtain the chat output message list in these Skype versions.

globals
; The main conversations window.
	handle SkypeWatch__appwin__conv5___hwndConv,
; A flag indicating whether this window has been found to be compatible with this module.
; -1 means no, 1 means yes, 0 means don't know yet.
	int SkypeWatch__appwin__conv5___compat,
; Cached list collections for live, contacts, conversations, and history.
; Key is a constant passed to SkypeWatch__appwin__conv5__listByName(),
; value is an MSAA object.
	collection SkypeWatch__appwin__conv5___listCache,
; RealName-to-scriptName map for the cache.
; Key is an actual MSAA list's name,
; value is the constant from this module used to find the MSAA object in the cache.
	collection SkypeWatch__appwin__conv5___listNames

int function SkypeWatch__appwin__conv5__applies()
; Returns True if this appears to be a Skype version supporting this module's interface.
var variant null
var handle hwnd = SkypeWatch__appwin__conv5___mainWindow()
if !hwnd
	SkypeWatch__appwin__conv5___listCache = null
	SkypeWatch__appwin__conv5___listNames = null
	return False
endIf
if hwnd != SkypeWatch__appwin__conv5___hwndConv
	SkypeWatch__appwin__conv5___compat = 0
	SkypeWatch__appwin__conv5___listCache = null
	SkypeWatch__appwin__conv5___listNames = null
endIf
if SkypeWatch__appwin__conv5___compat == 1
	return True
elif SkypeWatch__appwin__conv5___compat == -1
	return False
endIf
var int childID
var object o = getObjectFromEvent(hwnd, -4, 0, childID)
if !o
	return False
endIf
SkypeWatch__appwin__conv5___hwndConv = hwnd
var int n = o.accChildCount
if n > 20
	; That's permanent; we know this isn't compatible.
	SkypeWatch__appwin__conv5___compat = -1
	return False
endIf
childID = 1
while childID <= n
	if o.accRole(childID) == Role_System_List
		SkypeWatch__appwin__conv5___compat = 1
		return True
	endIf
	childID = childID +1
endWhile
; Don't mark permanently incompatible,
; in case this just means no list has yet been formed.
return False
endFunction

handle function SkypeWatch__appwin__conv5___mainWindow()
; Return the main conversations control window.
var handle hwnd = SkypeWatch__appwin__getSkypeWindow(appwin__SWT_Main)
if !hwnd
	return
endIf
hwnd = findWindow(hwnd, "TConversationsControl", "")
if !hwnd
	return
endIf
return hwnd
endFunction

object function SkypeWatch__appwin__conv5___mainClient()
; Return the main conversations control MSAA client object.
var handle hwnd = SkypeWatch__appwin__conv5___mainWindow()
if !hwnd
	return
endIf
var int childID
return getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
endFunction

string function SkypeWatch__appwin__conv5__peopleOnline()
; Return the string indicating how many people are online, if it can be found.
; The algorithm used is intended to be language independent.
if !SkypeWatch__appwin__conv5__applies()
	return ""
endIf
var object o = SkypeWatch__appwin__conv5___mainClient()
if !o
	return ""
endIf
var int childID = 1
var int n = o.accChildCount
if n > 20
	; Too much here to be right for Skype 5.
	return ""
endIf
while childID <= n
	if o.accRole(childID) == Role_System_StaticText
	&& SkypeWatch__appwin__conv5___isPeopleCountString(o.accName(childID))
		return o.accName(childID)
	endIf
	childID = childID +1
endWhile
return ""
endFunction

int function SkypeWatch__appwin__conv5___isPeopleCountString(string s)
; Returns True if s appears to represent the number of people online in Skype.
; The algorithm used is intended to be language independent.
var string digs = "0123456789"
if !stringContainsChars(s, digs)
	return False
endIf
var int i = 1
var int n = stringLength(s)
var int cnt = 0
var string ch
while i <= n
	ch = substring(s, i, 1)
	if stringContains(digs, ch)
		cnt = cnt +1
		if cnt > 4
			return True
		endIf
	else
		if cnt > 0 && !stringContains(digs, substring(s, i+1, 1))
			; Two separate numbers.
			return False
		endIf
	endIf
	i = i +1
endWhile
return False
endFunction

object function SkypeWatch__appwin__conv5__listByName(string listname)
; Return the list of given name or null.
if !SkypeWatch__appwin__conv5__applies()
	return
elif collectionItemExists(SkypeWatch__appwin__conv5___listCache, listname)
	return SkypeWatch__appwin__conv5___listCache[listname]
endIf
var object o = SkypeWatch__appwin__conv5___mainClient()
if !o
	return
endIf
var int childID = 1
var int n = o.accChildCount
while childID <= n
	if o.accRole(childID) == Role_System_List
	&& SkypeWatch__appwin__conv5___nameMatches(o.accName(childID), listname)
		if !SkypeWatch__appwin__conv5___listNames
			SkypeWatch__appwin__conv5___listCache = new collection
			SkypeWatch__appwin__conv5___listNames = new collection
		endIf
		var object oList = o.accChild(childID)
		listname = o.accName(childID)
		; Warning: Case is not significant here though it was before
		; use of JAWS 12+ collections.
		if collectionItemExists(SkypeWatch__appwin__conv5___listCache, listname)
			oList = SkypeWatch__appwin__conv5___listCache[listname]
		else
			SkypeWatch__appwin__conv5___listCache[listname] = oList
			SkypeWatch__appwin__conv5___listNames[o.accName(childID)] = listname
		endIf
		return oList
	endIf
	childID = childID +1
endWhile
return
endFunction

int function SkypeWatch__appwin__conv5___nameMatches(string name, string choices)
; Returns true if name is or is in choices.
if stringCompare(name, choices, False) == 0
	return True
elif stringContains("|"+choices+"|", "|"+name+"|")
	return True
endIf
return False
endFunction

string function SkypeWatch__appwin__conv5__focusedListName()
; Return the name of the focused list, or null if there isn't one.
var object o, int childID
o = getFocusObject(childID)
if !o
	return ""
endIf
var string name = o.accName(0)
if !name
	return ""
endIf
return SkypeWatch__appwin__conv5___listNames[name]
endFunction

string function SkypeWatch__appwin__conv5__focusShowingList()
; Focus the list that is already showing on screen.
; Returns True on success and False on failure.
var object o, int childID
o = SkypeWatch__appwin__conv5___mainClient()
if !o
	return False
endIf
var int n = o.accChildCount
childID = n
while childID
	if o.accRole(childID) == Role_System_List
		o.accSelect(SelFlag_TakeFocus, childID)
		return True
	endIf
	childID = childID -1
endWhile
return False
endFunction

object function SkypeWatch__appwin__conv5__chatContent(handle hwndChatOutput)
; Get the MSAA list object of chat messages from the given chat output window handle.
if !hwndChatOutput
	return
elif !SkypeWatch__appwin__conv5__applies()
	return
endIf
var int childID
var object o = getObjectFromEvent(hwndChatOutput, ObjID_Client, 0, childID)
if !o
	return
elif SkypeWatch__appwin__conv5___nameMatches(o.accName(1), appwin__conv5__ListChatOutput)
	return o.accChild(1)
elif SkypeWatch__appwin__conv5___nameMatches(o.accName(2), appwin__conv5__ListChatOutput)
	return o.accChild(2)
elif SkypeWatch__appwin__conv5___nameMatches(o.accName(3), appwin__conv5__ListChatOutput)
	return o.accChild(3)
endIf
return
endFunction

;------------------------------- End module conv5 ------------------------------

;============================== End module appwin ==============================


;================================== Module log =================================
; Allow logging to a text file.
;
; Interface:  SkypeWatch__log__open, SkypeWatch__log__isOpen, SkypeWatch__log__close, SkypeWatch__log__write, SkypeWatch__log__setUserID.
; Basic usage: SkypeWatch__log__open() from AutoStartEvent, SkypeWatch__log__write("...") to log something.
; Usage:
;	- Call SkypeWatch__log__open() from AutoStartEvent.
;	- Logging will occur by default only if the log file already exists (good for debugging during development).
;	- Call as SkypeWatch__log__open("", True) if you want logging to occur even if the log file doesn't already exist.
;	- Pass a file name instead of "" for a custom file name (rarely needed).
;	- Call SkypeWatch__log__write("Oops I saw an error") to log something.
;	- Use SkypeWatch__log__write("Oops this error needs to speak too", True) to speak what is being logged.
;	- Call SkypeWatch__log__close() to disable logging (rarely needed).
;
; Author:  Doug Lee of Level Access

; Constants for SkypeWatch__log__write and related functions.
; Not localized because log text is not localized.

; Used as a user ID when none is found.
const log___msgNoName = "<no name>"
; What "name" is used for the caller string when JAWS is too old for getScriptCallStack().
; %1 is the JAWS version.
const log___msgJAWSTooOld = "(caller unknown; JAWS version (%1) too old)"
; Name used for the caller when it can't be determined for an unknown reason,
; probably a getScriptCallStack return value format surprise.
const log___msgNoCaller = "(caller unknown)"

globals
; FileSystemObject and log stream object obtained from it.
	object SkypeWatch__log___fso,
	object SkypeWatch__log___logstream,
; Name of file.
	string SkypeWatch__log___fileName,
	int SkypeWatch__log___canCreate,
	string SkypeWatch__log___userID,
	string SkypeWatch__log___machineID

int function SkypeWatch__log___setup()
; Internal function called to set things up when necessary before any log text is generated.
if !stringIsBlank(SkypeWatch__log___userID)
	return True
endIf
SkypeWatch__log___fso = createObjectEx("Scripting.FileSystemObject", False)
if !SkypeWatch__log___fso
	return False
endIf
SkypeWatch__log___machineID = ""
var object oNet = createObjectEx("WScript.Network", False)
if oNet
	SkypeWatch__log___machineID = oNet.ComputerName
endIf
if stringIsBlank(SkypeWatch__log___machineID)
	SkypeWatch__log___machineID = log___msgNoName
endIf
SkypeWatch__log___userID = getJAWSUserName()
if stringIsBlank(SkypeWatch__log___userID) && oNet
	SkypeWatch__log___userID = oNet.UserName
endIf
if stringIsBlank(SkypeWatch__log___userID)
	SkypeWatch__log___userID = log___msgNoName
endIf
return True
endFunction

int function SkypeWatch__log__setUserID(string id)
; Optional means of setting a custom user ID.
; If not called, the value of getJAWSUserName() is used.
SkypeWatch__log___userID = id
return True
endFunction

int function SkypeWatch__log__close()
; Close the log file.
SkypeWatch__log___close()
SkypeWatch__log___fileName = ""
return True
endFunction

int function SkypeWatch__log___setFileName(string name)
; Set the log file name.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Helper for SkypeWatch__log__open().
var string basename
basename = name
if stringRight(basename,4) == ".log"
	basename = stringChopRight(basename, 4)
endIf
if !basename
	basename = getActiveConfiguration()
endIf
if !basename
	basename = getAppFileName()
	var int pos = stringContainsFromRight(basename, ".")
	if pos > 0
		basename = stringLeft(basename, pos-1)
	endIf
endIf
SkypeWatch__log___fileName = getJAWSSettingsDirectory() +"\\" +basename
if stringRight(SkypeWatch__log___fileName, 4) != ".log"
	SkypeWatch__log___fileName = SkypeWatch__log___fileName +".log"
endIf
return True
endFunction

int function SkypeWatch__log__open(/*?*/ string name, int canCreate)
; Open and optionally create the log file.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Call from autoStartEvent, or logging will not occur.
; Returns True if the file is opened and False if not.
if !SkypeWatch__log___setup()
	return False
endIf
SkypeWatch__log___canCreate = canCreate
SkypeWatch__log__close()
if !SkypeWatch__log___setFileName(name) || !SkypeWatch__log___open()
	SkypeWatch__log___fileName = ""
	return False
endIf
return True
endFunction

int function SkypeWatch__log___open()
; Raw open/create.  Sets SkypeWatch__log___logstream.
; Returns True on success and False on failure.
if SkypeWatch__log___logstream
	return True
endIf
if fileExists(SkypeWatch__log___fileName)
	SkypeWatch__log___logstream = SkypeWatch__log___fso.openTextFile(SkypeWatch__log___fileName, 8)  ; Append
elif SkypeWatch__log___canCreate
	SkypeWatch__log___logstream = SkypeWatch__log___fso.CreateTextFile(SkypeWatch__log___fileName)
endIf
if !SkypeWatch__log___logstream
	return False
endIf
scheduleFunction("SkypeWatch__log___close", 30)
return True
endFunction

int function SkypeWatch__log___close()
; Raw cloase; unsets SkypeWatch__log___logstream.
var variant null
if !SkypeWatch__log___logstream
	return True
endIf
SkypeWatch__log___logstream.Close()
SkypeWatch__log___logstream = null
return True
endFunction

int function SkypeWatch__log___isOpen()
; Returns True if the log file is actually open and False if not.
if SkypeWatch__log___logstream
	return True
endIf
return False
endFunction

int function SkypeWatch__log__isOpen()
; Returns True if the log file is open and False if not.
; "Open" here means can be written to but may not necessarily mean the file is physically opened at the moment.
if SkypeWatch__log___fileName
	return True
endIf
return False
endFunction

string function SkypeWatch__log___getStamp()
; Returns a formatted date/time stamp for the log file.
var string tm = sysGetTime("HH:mm:ss")
var string dt = sysGetDate("ddd MMM dd yyyy")
return formatString("%1 %2", tm, dt)
endFunction

int function SkypeWatch__log__write(string s, /*?*/ int alsoSpeak)
; Write a string to the log file, and also optionally speak it.
if alsoSpeak
	sayMessage(OT_Error, s)
endIf
if !SkypeWatch__log__isOpen()
	return False
endIf
if !SkypeWatch__log___open()
	return False
endIf
var string buf = formatString(
	"*** %1 %2@%3 (%4)\n%5:\n",
	SkypeWatch__log___getStamp(), SkypeWatch__log___userID, SkypeWatch__log___machineID, getAppFileName(),
	SkypeWatch__log___getCaller()
) +s +"\n\n"
; Standardize any CR/LF characters in buf before writing.
buf = stringReplaceChars(buf, "\r\n", "\n")
buf = stringReplaceSubstrings(buf, "\n", "\r\n")
SkypeWatch__log___logstream.write(buf)
return True
endFunction

string function SkypeWatch__log___getCaller()
; Get and return the caller of this logging system.
; Helper for SkypeWatch__log__write().
if getJFWVersion() < 1000000
	var string jver = stringSegment(SkypeWatch__util__getJAWSInfo(), "|", 1)
	return formatString(log___msgJAWSTooOld, jver)
endIf
var string stack = SkypeWatch__util__callStackInfo(2, "s", 2)
if !stringIsBlank(stack)
	return stack
endIf
return formatString("%1\nStack begin:\n%2\nStack end.",
	log___msgNoCaller,
	SkypeWatch__util__vcast(getScriptCallStack())
)
endFunction

string function SkypeWatch__log___splitOff(string byRef s, string sep, /*?*/ int fromRight)
; Split what's before sep off and return it.
; If fromRight is True, split everything before the last sep off and return that.
; Returns by reference s with both that and sep removed.
var string s1
var int pos
if fromRight
	pos = stringContainsFromRight(s, sep)
else
	pos = stringContains(s, sep)
endIf
if pos
	s1 = stringLeft(s, pos-1)
	s = stringChopLeft(s, pos+stringLength(sep)-1)
else
	s1 = s
	s = ""
endIf
return s1
endFunction

;================================ End module log ===============================


;========================= Module supdate (from update) ========================
; This code causes these scripts to arrange for the presence of <filename>_fs.jsb/jsd (copied from <filename>.jsb/jsd) in the user settings directory.
; They are automatically obtained from the shared settings directory if not already found in the user settings directory.
; <filename> can be any jsb file (extension not included) or set of files.
; This code uses the Scripting.FileSystemObject and requires JAWS 6.0 or later.
; No file copying is performed under JAWS 14 or later, because a better means exists to solve this problem there.
;
; Interface: Call SkypeWatch__supdate__checkFiles() from autoStartEvent, passing a delimiter and one or more file basenames to keep in sync.
; Example: SkypeWatch__supdate__checkFiles("|default").
;
; Author:  Doug Lee

globals
; The files to update as needed.
; Format of string: <sep>fname1[<sep>fname2...], where <sep> is the separator, e.g., |.
	string SkypeWatch__supdate___flist,
; A flag to keep this module's ReloadAllConfigs() call from causing an infinite loop if an update fails.
	int SkypeWatch__supdate___internal

int function SkypeWatch__supdate___getFSObject(object byRef o)
var
	string objname
objname = "scripting.fileSystemObject"
if o
	return True
endIf
o = createObjectEx(objname, False)
if !o
	;sayMessage(OT_Error, formatString(supdate___msgCantCreateObject, objname))
	return False
endIf
return True
endFunction

int function SkypeWatch__supdate___deleteUserFile(object o, string fname)
; Deletes a file from the user settings folder (if it exists) and returns True if the file is gone on exit.
; Read-only files are deleted by force here.
if !stringContains(fname, ":")
	fname = getUserSettingsDirectory() +"\\" +fname
endIf
if !fileExists(fname)
	return True
endIf
o.deleteFile(fname, True)
if fileExists(fname)
	sayMessage(OT_Error, formatString(supdate___msgCantDeleteFile, fname))
	return False
endIf
return True
endFunction

int function SkypeWatch__supdate___copySharedFile(object o, string src, string dest)
; Copies a file of name src from the shared settings folder to a file of name dest in the user settings directory.
; If dest is null, the src name is also used for the destination file name.
; Returns True on success and False on failure.
; If dest exists before copy, it is first deleted, even if it is read-only.
; Note that if the copy fails, any original dest file is not recovered.
if !stringLength(dest)
	dest = src +""
endIf
if !stringContains(src, ":")
	src = getSharedSettingsDirectory() +"\\" +src
endIf
if !stringContains(dest, ":")
	dest = getUserSettingsDirectory() +"\\" +dest
endIf
if fileExists(dest)
	if !SkypeWatch__supdate___deleteUserFile(o, dest)
		sayMessage(OT_Error, formatString(supdate___msgCopyFailed, src))
		return False
	endIf
endIf
o.copyFile(src, dest, True)
if !fileExists(dest)
	sayMessage(OT_Error, formatString(supdate___msgCopyFailed, src))
	return False
endIf
return True
endFunction

void function SkypeWatch__supdate___requestJAWSRestart()
; Call via ScheduleFunction if doing it directly doesn't produce the box for some reason.  [DGL, 2006-01-24]
messageBox(supdate___msgChangeNotice)
endFunction

int function SkypeWatch__supdate___updateFile(string dest, string src, int force)
; Updates src if dest seems to have changed since this was last done, or if force is True.
var
	object oFS, object oDest, object oSrc,
	int update,
	string name
update = False
if !fileExists(src)
	sayMessage(OT_Error, formatString(supdate___msgSourceFileNotFound, src))
	return False
endIf
if !SkypeWatch__supdate___getFSObject(oFS)
	return False
endIf
oSrc = oFS.getFile(src)
if !oSrc
	return False
endIf
oDest = oFS.getFile(dest)
if !oDest
	update = True
elif !stringCompare(oSrc.dateLastModified, oDest.dateLastModified, True) && oSrc.size == oDest.size
	if !force
		return False
	endIf
	update = True
else
	update = True
endIf
if !update
	return False
endIf
name = stringChopLeft(dest, stringContainsFromRight(dest, "\\"))
sayUsingVoice(VCTX_Message, formatString(supdate___msgUpdatingFile, name), OT_Status)
SkypeWatch__supdate___copySharedFile(oFS, oSrc.name, name)
return True  ; Assume there was a change even on copy failure.
endFunction

int function SkypeWatch__supdate___inHJAlert()
; Like inHJDialog() but only returns True on dialogs that appear to be alerts or confirmations.
var
	handle hwnd,
	int typeCode
if !inHJDialog()
	return False
endIf
hwnd = getFirstChild(getTopLevelWindow(getFocus()))
; Only have to scan the first level; buttons and statics should not have descendants here,
; and anything else will throw False back to the caller anyway.
while hwnd
	typeCode = getWindowTypeCode(hwnd)
	if typeCode != WT_Static && typeCode != WT_Button
		return False
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
; Statics, buttons, and nothing else; looks like an alert to me.
return True
endFunction

int function SkypeWatch__supdate__checkFiles(string flist)
; Sets up the required files for these scripts.
; Requires JAWS 6.0 or later.
if SkypeWatch__supdate___internal
	; This call resulted from a ReloadAllConfigs() call from this module,
	; so in order to avoid an infinite loop, we'll do nothing here.
	; (This happens if an update fails, such as because of a read-only destination file.)
	SkypeWatch__supdate___internal = False
	return
elif getJFWVersion() >= 1400000
	; Do nothing because JAWS 14 and later need not make these copies.
	return
endIf
SkypeWatch__supdate___flist = flist
scheduleFunction("SkypeWatch__supdate___checkFiles0", 10)
endFunction

void function SkypeWatch__supdate___checkFiles0()
; Does the bulk of the work of SkypeWatch__supdate__checkFiles().
; Done indirectly because I saw direct (unscheduled) execution of similar code cause a JAWS and IE crash once.  [DGL, 2006-01-24]
; This also allows any JAWS "script not found" dialogs to appear before we try to dismiss them.
var
	object o,
	string sharedPath, string userPath,
	int force,
	string sep, string fname,
	string src, string dest,
	int i, int n, int changed

; First dismiss the error box that shows up when we try loading a jsb file with a "Use" line when that file doesn't exist yet.
; These scripts abort the announcement of that error, but the user will still hear a Ding sound from this dialog's appearance.
if SkypeWatch__supdate___inHJAlert()
	stopSpeech()
	typeKey("Escape")
	pause()
	stopSpeech()
endIf

if !SkypeWatch__supdate___getFSObject(o)
	return
endIf
sharedPath = getSharedSettingsDirectory()
userPath = getUserSettingsDirectory()
force = False

sep = stringLeft(SkypeWatch__supdate___flist, 1)
n = stringSegmentCount(SkypeWatch__supdate___flist, sep)
i = 2
changed = False
while i <= n
	fname = stringSegment(SkypeWatch__supdate___flist, sep, i)
	src = sharedPath +"\\" +fname
	dest = userPath +"\\" +fname +"_fs"
	changed = SkypeWatch__supdate___updateFile(dest+".jsb", src+".jsb", force) ||changed
	changed = SkypeWatch__supdate___updateFile(dest+".jsd", src+".jsd", force) ||changed
	i = i +1
endWhile
if changed
	SkypeWatch__supdate___internal = True
	ReloadAllConfigs()
	; This triggers an AutoStartEvent call in this script file.
	; We don't want that call to trigger another checkFiles call,
	; or an update failure, such as caused by a read-only destination file,
	; would cause an infinite check/update/check/... loop.
	; The AutoStartEvent call happens after the ReloadAllConfigs() call returns though,
	; so we have to schedule the reset of SkypeWatch__supdate___internal rather than just clearing it here.
	scheduleFunction("SkypeWatch__supdate___clearInternal", 50)
	sayUsingVoice(VCTX_Message, supdate___msgDone, OT_Status)
else
	;sayUsingVoice(VCTX_Message, supdate___msgNothingToUpdate, OT_Status)
endIf
endFunction

void function SkypeWatch__supdate___clearInternal()
; Clear SkypeWatch__supdate___internal after a ReloadAllConfigs() call from this module.
; See comments in SkypeWatch__supdate___checkFiles0() for details.
SkypeWatch__supdate___internal = False
endFunction

int function SkypeWatch__supdate__deconfigure()
; Do what is necessary to make the next load of these scripts reconfigure things as if installed anew.
; This function is not called by anything but can be called by test code while the self-configuration system is being tested.
; Returns True on success and False if anything failed.
var
	object o,
	string sep, string fname,
	int i, int n, int failed
if !SkypeWatch__supdate___getFSObject(o)
	return False
endIf
sep = stringLeft(SkypeWatch__supdate___flist, 1)
n = stringSegmentCount(SkypeWatch__supdate___flist, sep)
i = 2
failed = False
while i <= n
	fname = stringSegment(SkypeWatch__supdate___flist, sep, i)
	failed = (!SkypeWatch__supdate___deleteUserFile(o, fname+"_fs.jsb")) || failed
	failed = (!SkypeWatch__supdate___deleteUserFile(o, fname+"_fs.jsd")) || failed
	i = i +1
endWhile
return !failed
endFunction

;======================= End module supdate (from update) ======================


;================================= Module xlist ================================
; XList, the arbitrary list (array and hash emulation) module.
;
; Terminology:
;	XList - An arbitrary list of items (an array, basically).
;	XLHash - An XList in which odd-numbered items are keys and even-numbered items are their values.
;
; Implementation Notes:
; XLists always begin and end with a delimiter; a null list is just a single delimiter character.
; This means it is NOT possible to use boolean (if sl) or length (if stringLength(sl)) to check for emptiness; use isEmpty or itemCount.
; Objects can not be stored in an XList directly, and attempting this will cause an error message to speak.
;
; Author:  Doug Lee
;
; Revision History:  See version control logs.

const
; Indexes used to request the character tables supported by this module.
	xlist___CharANSI = 1,
; The corresponding character tables.
; The double quote (") and backslash characters must be escaped.
	xlist___CharANSICharList = "\1\2\3\4\5\6\7\8\9\10\11\12\13\14\15\16\17\18\19\20\21\22\23\24\25\26\27\28\29\30\31 !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\127\128\129\130\131\132\133\134\135\136\137\138\139\140\141\142\143\144\145\146\147\148\149\150\151\152\153\154\155\156\157\158\159\160\161\162\163\164\165\166\167\168\169\170\171\172\173\174\175\176\177\178\179\180\181\182\183\184\185\186\187\188\189\190\191\192\193\194\195\196\197\198\199\200\201\202\203\204\205\206\207\208\209\210\211\212\213\214\215\216\217\218\219\220\221\222\223\224\225\226\227\228\229\230\231\232\233\234\235\236\237\238\239\240\241\242\243\244\245\246\247\248\249\250\251\252\253\254\255"

string function SkypeWatch__xlist__toJAWSList(string lst, string sep)
; Returns lst restructured as a "normal" JAWS list (no leading and trailing separators) with the given delimiter.
; If the list can't be converted so, the null string is returned.
; This happens when the given delimiter is already part of the content of the list.
; Multi-character delimiters are supported here, as is a null delimiter.
var
	string delim,
	int len
delim = stringLeft(lst, 1)
if stringCompare(sep, delim, True)
	if stringContains(lst, sep)
		return False
	endIf
	SkypeWatch__xlist___setDelim(lst, sep, 1)
endIf
len = stringLength(sep)
return stringChopLeft(stringChopRight(lst, len), len)
endFunction

string function SkypeWatch__xlist__fromJAWSList(string lst, string sep)
; Builds an XList from the given "normal" JAWS list with given separator.
; The separator may be multi-character.
var
	int len
len = stringLength(sep)
lst = SkypeWatch__string__cat(sep, SkypeWatch__string__cat(lst, sep))
SkypeWatch__xlist___settleDelim(lst, "", len)
return lst
endFunction

int function SkypeWatch__xlist__resize(string byRef lst, int n)
; The externally visible version of SkypeWatch__xlist___resize().
var
	int n0
n0 = SkypeWatch__xlist__itemCount(lst)
return SkypeWatch__xlist___resize(lst, n0, n)
endFunction

int function SkypeWatch__xlist__insertItem(string byRef lst, int i, variant item)
; Inserts item before the ith element, making item become element i.
var
	int pos
if !stringLength(item) && item
	sayMessage(OT_No_Disable, xlist___msgNoObjects)
	return False
endIf
if !SkypeWatch__xlist___settleDelim(lst, item, 1)
	return False
endIf
pos = SkypeWatch__xlist___indexPos(lst, i)
if !pos
	return False
endIf
lst = SkypeWatch__string__cat(stringLeft(lst, pos-1), SkypeWatch__string__cat(item, stringChopLeft(lst, pos-2)))
return True
endFunction

int function SkypeWatch__xlist___resize(string byRef lst, int n0, int n)
; Grow or shrink lst to n elements.
; Growing lst amounts to appending null elements; shrinking amounts to chopping elements off the end.
; This internal function expects to be told the current element count; this is for efficiency.
if !stringLength(lst)
	lst = SkypeWatch__xlist__new()
endIf
if n < 0
	return False
elif n == 0
	lst = stringLeft(lst, 1)
	return True
elif n == n0
	return True
elif n == n0-1
	lst = stringChopRight(lst, stringLength(SkypeWatch__xlist__getItem(lst, n+1))+1)
	return True
elif n < n0
	lst = stringLeft(lst, SkypeWatch__xlist___indexPos(lst, n+1))
	return True
else  ; n > n0
	lst = SkypeWatch__string__cat(lst, SkypeWatch__string__repeat(stringLeft(lst, 1), n-n0))
	return True
endIf
return False  ; never reached
endFunction

variant function SkypeWatch__xlist__getItem(string lst, int i)
; Returns the ith element of lst.
; If i is greater than the item count, the null string is returned.
var string val
if i < 1 || !stringLength(lst)
	return ""
endIf
val = stringSegment(lst, stringLeft(lst, 1), i+1)
return val
endFunction

int function SkypeWatch__xlist__itemCount(string lst)
; Returns the number of items in lst.
; If lst is not a proper XList, returns 0.
if !stringLength(lst)
	return 0
endIf
return SkypeWatch__string__segmentCount(lst, stringLeft(lst, 1)) -2
endFunction

int function SkypeWatch__xlist___settleDelim(string byRef lst, string item, int dlen)
; Internal function to manage the changing of delimiters as needed.
; Must be called whenever material is being added to an XList.
; dlen is the length of lst's current delimiter (1 except when sent via fromJAWSList).
; Returns True if a valid delimiter is in place on exit and False if there is no safe delimiter available.
; A "valid" delimiter is a one-character delimiter not found in lst (except as the current delimiter) or item.
var
	string delim,
	int c0, int c, string ch,
	int done
if !stringLength(lst)
	lst = SkypeWatch__xlist__new()
endIf
delim = stringLeft(lst, dlen)
if stringContains(item, delim) || dlen != 1
	c = SkypeWatch__xlist__charToANSI(stringLeft(lst, 1))
	c0 = c
	done = False
	while !done
		c = c +1
		if c > 254
			; 255 may not write correctly to a file, and 0 is not possible in JAWS.
			c = 1
		elif c >= 8 && c <= 13
			; backspace, tab, LF, RF(?), FF, CR:
			; These characters are unfriendly when written to a text file, so they are avoided.
			c = 14
		elif c == 26 || c == 27
			; DOS/Windows EOF, ESC
			c = 28
		endIf
		ch = SkypeWatch__xlist__charFromANSI(c)
		if c == c0
			; We've come all the way around; there is no safe delimiter for this situation.
			beep()
			sayMessage(OT_No_Disable, xlist___msgOutOfDelims)
			return False
		elif !stringContains(lst, ch) && !stringContains(item, ch)
			; We have a winner!
			SkypeWatch__xlist___setDelim(lst, ch, dlen)
			delim = ch
			done = True
		endIf
	endWhile
endIf
return True
endFunction

int function SkypeWatch__xlist__charToANSI(string c)
; Returns the ANSI code for the first character in c, or 0 if c is null.
return SkypeWatch__xlist___getCharCode(c, xlist___CharANSI)
endFunction

string function SkypeWatch__xlist__charFromANSI(int i)
; Returns the character corresponding to the given ANSI code.
return SkypeWatch__xlist___getChar(i, xlist___CharANSI)
endFunction

string function SkypeWatch__xlist__new()
; Create and return a new XList.
return "\1"
endFunction

int function SkypeWatch__xlist___indexPos(string lst, int i)
; Returns the character position in lst of the first character of the ith element.
; Returns 0 if there is no such element.
; Exception:  If i == itemCount+1, returns stringLength(lst)+1.
var
	string delim,
	int len,
	int pos,
	int result
if i < 1
	return 0
endIf
delim = stringLeft(lst, 1)
len = stringLength(lst)
if !len
	return 0
elif len == 1
	return 1
endIf
result = 0
while i
	pos = stringContains(lst, delim)
	if !pos
		return 0
	endIf
	result = result +pos
	lst = stringChopLeft(lst, pos)
	i = i -1
endWhile
return result +1
endFunction

int function SkypeWatch__xlist___getCharCode(string s, int whichList)
; Converts the first character in s to the corresponding code based on the indicated character table.
; If s is null, 0 is returned.
; Can't handle \000 because this seems to end JAWS strings.
; In fact, including \000 anywhere in a string causes it to appear empty!
; [DGL, 02/28/03]
s = stringLeft(s, 1)  ; in case s contains more than one character
if !stringLength(s)
	return 0
endIf
return stringContains(SkypeWatch__xlist___getCharList(whichList), s)
endFunction

string function SkypeWatch__xlist___getChar(int i, int whichList)
; Converts the given code to the corresponding character from the indicated character table.
; If i is less than 1 or above the length of the table, the null string is returned.
var string charList
if i <= 0
	return ""
endIf
charList = SkypeWatch__xlist___getCharList(whichList)
if i > stringLength(charList)
	return ""
endIf
return substring(charList, i, 1)
endFunction

string function SkypeWatch__xlist___getCharList(int whichList)
; Returns the requested character list.  This is to avoid duplicating huge constants in the script binary.
if whichList == xlist___CharANSI
	return xlist___CharANSICharList
endIf
return ""
endFunction

int function SkypeWatch__xlist___setDelim(string byRef lst, string ch, int dlen)
; Sets lst's delimiters to ch.
; dlen is the length of lst's current delimiter (1 except when sent via fromJAWSList).
var
	string delim
delim = stringLeft(lst, dlen)
if stringLength(ch) == 1 && dlen == 1
	lst = SkypeWatch__string__ref(stringReplaceChars(lst, delim, ch))
else
	; {To/From}JAWSList() can send a multi-character delimiter here, but no proper XList will use one.
	lst = SkypeWatch__string__ref(stringReplaceSubstrings(lst, delim, ch))
endIf
return True
endFunction

;=============================== End module xlist ==============================


;================================= Module util =================================
; Miscellaneous utility functions
;
; Author:  Doug Lee

string function SkypeWatch__util__stringFromFile(string fname, /*?*/ string fmt)
; Returns the file contents as a string.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt,
	string s
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 1, 0, iFmt)  ; 1 is ForReading, 0 means don't create
if !f
	;sayString("No code file")
	return ""
endIf
s = f.readAll()
f.close()
if !stringLength(s)
	;sayString("No code in the file")
endIf
return s
endFunction

int function SkypeWatch__util__stringToFile(string fname, string s, /*?*/ string fmt)
; Writes the string to the indicated file, creating or replacing the file.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 2, 1, iFmt)  ; 2 is ForWriting, 1 means create
if !f
	;sayString("Can't create file")
	return ""
endIf
f.write(s)
f.close()
return True
endFunction

handle function SkypeWatch__util__getFirstTopLevelWindow()
; Returns the handle of the first window at the top level of the window tree.
var
	handle hwnd

hwnd = SkypeWatch__util__getDesktopWindow()
if hwnd
	return getFirstChild(hwnd)
endIf

; Start at the focused window, or if there isn't one, from an arbitrary window on the screen.
hwnd = getFocus()
if !hwnd
	hwnd = getWindowAtPoint(300, 300)  ; A valid point regardless of screen resolution
endIf

; Find the top level of the window tree without relying on getAppMainWindow.
; Note that this can actually reach the parent of the top level if we are starting on a ComboLBox window!
while getParent(hwnd)
	hwnd = getParent(hwnd)
endWhile

; make sure this is the level we want, not the desktop itself.
; At least in Windows 95/98, the desktop has window type 14 (desktop), class "#32769", and handle 128.
; The handle is 65566 rather than 128 in Windows NT.  [DGL, 10/07/01]
; We use the first of the above properties because it seems the most likely to remain unique for a window with no parents.
if !getNextWindow(hwnd) && !getPriorWindow(hwnd)
	hwnd = getFirstChild(hwnd)
else
	hwnd = getFirstWindow(hwnd)
endIf
return hwnd
endFunction

handle function SkypeWatch__util__getDesktopWindow()
; Analogous to the Windows call of the same name.
var
	handle hwnd,
	string hlist, int hcount, int i

; Switching to the invisible cursor keeps GetWindowTypeCode from returning 3 (WT_Edit) for all windows when the virtual cursor is active.
saveCursor() invisibleCursor()

; First try handles known to be the Desktop window in various versions of Windows.
; 128=95/98, 65566=NT, 65556=XP, 65546=2k.
; Thanks to Victor Tsaran for helping collect these handles. [DGL, 04/28/03]
; 65552 for Windows 8.1.
hlist = "128|65566|65556|65552|65546|65536|65526|65516|65506"
hcount = 8
i = 1
while i <= hcount
	hwnd = stringToHandle(stringSegment(hlist, "|", i))
	if getWindowTypeCode(hwnd) == WT_Desktop
		return hwnd
	endIf
	i = i +1
endWhile

; The hard way...

; Start at the focused window, or if there isn't one, from an arbitrary window on the screen.
hwnd = getFocus()
if !hwnd
	hwnd = getWindowAtPoint(300, 300)  ; A valid point regardless of screen resolution
endIf
hwnd = getFirstWindow(getTopLevelWindow(hwnd))
while hwnd
	if getWindowTypeCode(getParent(hwnd)) == WT_Desktop
		return getParent(hwnd)
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

string function SkypeWatch__util__callStackInfo(int nlevels, string flags, /*?*/ int skip)
; Produce a representation of the current call stack formatted according to the given parameters.
; The following shortenings are applied to each script/function call in the return value:
;	- The shared folder path is replaced with "shared."
;	- The user folder path is removed.
;	- The word "function" is removed.
;	- The word "script" is replaced with "$."
; Requires JAWS 10 or later because the JAWS 10+ getScriptCallStack() is required.
; nlevels: Number of levels from starting position up the stack to print.  1 means just current function.  0 or less means everything.
; flags: f for forward order (default is reverse), s for short (one line and no parameters).
; skip: How many calls to skip over besides this function itself, default 0.
; Returns the requested call stack information or null if the JAWS version is too old.
var
	string userFolder, string sharedFolder,
	int short, int forward,
	int stacksize
if getJFWVersion() < 1000000
	return ""
endIf
flags = stringLower(flags)
short = stringContains(flags, "s")
forward = stringContains(flags, "f")
userFolder = getJAWSSettingsDirectory()
sharedFolder = getSharedSettingsDirectory()
var collection stack = SkypeWatch__util__callStack()
var int i = 2  ; skip this function's entry.
i = i +skip
stacksize = stack.count
if nlevels > 0
	nlevels = min(nlevels, stacksize-i+1)
else
	nlevels = stacksize-i+1
endIf
if nlevels <= 0
	return ""
endIf
var int delta = 1
if forward
	delta = -1
	i = i +nlevels -1
endIf
var string buf = ""
while nlevels
	if buf
		if forward
			buf = buf +" -> "
		else
			buf = buf +" <- "
		endIf
	endIf
	var collection  entry = stack[intToString(i)]
	var string fname = entry.file
	fname = stringReplaceSubstrings(fname, sharedFolder, "")
	fname = stringReplaceSubstrings(fname, userFolder, "user")
	if stringRight(fname, 4) == ".jsb"
		fname = stringChopRight(fname, 4)
	endIf
	buf = buf +fname +" "
	if entry.type == "script"
		buf = buf +"$"
	endIf
	if short
		buf = buf +stringSegment(entry.call, "(", 1)
	else
		buf = buf +SkypeWatch__string__ref(entry.call)
	endIf
	i = i +delta
	nlevels = nlevels -1
endWhile
return buf
endFunction

globals object SkypeWatch__util___csRE
collection function SkypeWatch__util__callStack()
; Return a collection of collections representing the script call stack.
; Requires JAWS 11 update 2 or later and the VBScript.RegExp object.
; Index the result by stringified 1-based integers.
; The most recent call is item 1.
; Structure of an item:
;	file: The file from which the script/function came, either "builtin" or a full path.
;	type: The type ("script" or "function", lower case).
;	call: The full call with parameters.
if !SkypeWatch__util___csRE
	SkypeWatch__util___csRE = SkypeWatch__util__createObject("VBScript.RegExp")
	if !SkypeWatch__util___csRE
		return
	endIf
	SkypeWatch__util___csRE.Global = -1
	SkypeWatch__util___csRE.IgnoreCase = 0
	SkypeWatch__util___csRE.Pattern = "\10([^\10\t]*)\t(script |function )"
endIf
var string sStack = "\10" +getScriptCallStack()
var object matches = SkypeWatch__util___csRE.execute(sStack)
var object match
var int n = matches.count
var int i = 1  ; Skip this function itself.
var string tmp, int pos1, int pos2
var collection stack stack = new collection
var collection entry
stack.count = 0
while i < n
	entry = new collection
	stack.count = stack.count +1
	stack[intToString(stack.count)] = entry
	match = matches(i)
	; Skip the \10 here.
	; Also switch from 0-base to 1-base.
	pos1 = match.firstIndex +2
	if i == n-1
		pos2 = stringLength(sStack) +1
	else
		pos2 = matches(i+1).firstIndex +1
	endIf
	tmp = substring(sStack, pos1, pos2-pos1)
	entry.file = SkypeWatch__string__segmentNext(tmp, "\t")
	entry.type = SkypeWatch__string__segmentNext(tmp, " ")
	entry.call = tmp
	i = i +1
endWhile
return stack
endFunction

globals object SkypeWatch__util___fso
object function SkypeWatch__util__getFSObject()
; Cached access to a filesystem object.
if !SkypeWatch__util___fso
	SkypeWatch__util___fso = SkypeWatch__util__createObject("Scripting.FilesystemObject")
endIf
return SkypeWatch__util___fso
endFunction

string function SkypeWatch__util__getJAWSInfo()
; Returns a list of fields, separated by "|", as follows:
; 1.  The JAWS version number from the JFWUI2 window, in the form [M]M.mm.bbb; example, 5.00.584.
; 2.  The JAWS serial number from the JFWUI2 window.
; 3.  The JAWS serial number from GetJFWSerialNumber().
; The build number returned by this function is accurate, whereas it is "000" in GetJFWVersion in older JAWS versions.
var
	handle hwnd0, handle hwnd,
	int pos,
	string sn, string ver

; Find the JAWS dialog, which exists even when invisible or when JAWS is on the system tray.
hwnd0 = getFirstChild(findTopLevelWindow("JFWUI2", ""))
if !hwnd0
	return ""
endIf

; Get the serial number from the status bar's window name.
hwnd = getNextWindow(hwnd0)
sn = stringLower(SkypeWatch__util__getFullWindowName(hwnd))
pos = stringContains(sn, "ber")
if pos
	sn = intToString(stringToInt(substring(sn, pos+4, 10)))
else
	sn = ""
endIf

; Find the static containing the version number and get it.
hwnd = getFirstChild(hwnd0)
ver = ""
while hwnd && !ver
	ver = stringTrimTrailingBlanks(SkypeWatch__util__getFullWindowName(hwnd))
	pos = stringContainsFromRight(ver, " ")
	if pos
		ver = stringChopLeft(ver, pos)
	else
		ver = ""
	endIf
	hwnd = getNextWindow(hwnd)
endWhile

; The above version retrieval sometimes fails, so try the normal way.
if !ver
	ver = intToString(getJFWVersion())
	var int verlen = stringLength(ver)
	ver = formatString("%1.%2.%3",
		stringChopRight(ver, 5),
		substring(ver, verlen-4, 1),
		stringRight(ver, 4)
	)
endIf

if !sn && !ver
	return ""
else
	return ver +"|" +sn +"|" +getJFWSerialNumber()
endIf
endFunction

string function SkypeWatch__util__getFullWindowName(handle hWnd)
; Gets the full name of the given window if possible, using MSAA when it seems advisable.
; The JAWS GetWindowName() function seems to cut the name at around 255 characters.
; JAWS can get up to 4 K (maybe more sometimes) from MSAA though.
; Some JAWS versions (15?) can also return null for getWindowName when MSAA can get it.
var
	string wName,
	string mName,
	int cutoffLen

; Window names can reach 255 characters, but we choose an earlier cutoff for safety.
cutoffLen = 240

; Zero-length names can mean getWindowName missed it entirely. [DGL, 2014-09-29, JAWS 15.0]
wName = GetWindowName(HWnd)
var int len = stringLength(wName)
if !len || len > cutoffLen
	var int childID, object o
	o = GetObjectFromEvent(HWnd, -4, 0, childID)  ; -4 = ObjID_Client
	mName = o.accName(0)
	if stringLength(mName) > len
		return mName
	endIf
endIf
return wName
EndFunction

variant function SkypeWatch__util__vcast(variant v)
; Cast any type to variant.
return v
endFunction

object function SkypeWatch__util__createObject(string objname)
; A CreateObject wrapper that avoids causing 5-7 second delays in DOS boxes when called if JAWS is running as a service.
; As of June 11, 2011, uses the CoCreateInstance method of creating an object
; unless that doesn't work, whereupon the GetObjectHelper method is used.
; Previously the CoCreateInstance method was only used if JAWS was running as a service.
var object o = createObjectEx(objname, False)
if !o
	o = createObject(objname)
endIf
return o
endFunction

;=============================== End module util ===============================


;================================ Module string ================================
; String manipulation functions.
;
; Author:  Doug Lee

string function SkypeWatch__string__segmentNext(string byRef s, string sep)
; Gets, returns, and removes from s the next string segment and the trailing separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
pos = 0
if sep
	pos = stringContains(s, sep)
endIf
if !pos || !sep
	s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	s = ""
	return s1
endIf
s1 = stringLeft(s, pos-1)
s = stringChopLeft(s, pos +stringLength(sep) -1)
return s1
endFunction

int function SkypeWatch__string__segmentCount(string s, string sep)
; Returns the number of segments in the given string.
; If sep is null, the result is 0.
; As of 03/15/03, the null string is defined to have zero segments (previously one).
; Any non-null string containing no separators has one segment.
; WARNING: Unlike the built-in (as of JAWS 7.0) stringSegmentCount function,
; this function counts the null string after a trailing delimiter;
; thus, SkypeWatch__string__segmentCount("a|b|", "|") is 3, where JAWS would say 2.
; Also, this function considers a multi-character sep as a multi-character delimiter,
; whereas stringSegmentCount allows each character of sep to be a delimiter.
; This function requires JAWS 7.0 or later.
var
	int len,
	int seplen,
	int l,
	int n
seplen = stringLength(sep)
if seplen == 1
	n = stringSegmentCount(s, sep)
	if stringRight(s, 1) == sep
		; JAWS does not count the null string after a trailing delimiter, but I do.
		n = n +1
	endIf
	return n
endIf
len = stringLength(s)
if !len
	return 0
endIf
if !seplen
	return 0
endIf

; A little optimization made possible by JAWS 6.0's StringReplaceSubstrings function...
; Count segments by removing all separators and comparing string lengths.
l = stringLength(SkypeWatch__string__ref(stringReplaceSubstrings(s, sep, "")))
l = len -l
return (l / seplen) +1
endFunction

int function SkypeWatch__string__replace(string byRef s, string r1, string r2, string whatToDo)
; Replace r1 in s with r2 if r1 is found where whatToDo indicates.
; Note:  A space is sometimes prepended to and then removed from results in this function to keep JAWS from adding
; numeric strings arithmetically, which can happen in JAWS versions at least through 3.70.47 [DGL, 12/20/2000]
; whatToDo: l=left, r=right, a=anywhere, e=everywhere.
; Add t (for Trim) if whitespace should be removed at the point where s is split.
; Add c (for ConsiderCase) if case-sensitive matching is desired.
; For compatibility with older scripts, WhatToDo can also be an integer sum of flags:
; 0=l, 1=r, 2=a, 3=e, 8=t, 16=c.
; Returns the number of changes made to s, which must be passed by reference.
; If s is null, no changes are made and 0 is returned.
; If r1 is null, the following happens based on WhatToDo:
;	"e" - r2 is inserted between each character pair in s.  All blanks are removed first if "t" is included.
;	"l" or "a" - r2 is inserted after the first character of s, after blanks around it are removed if "t" is included.
;	"r" - r2 is inserted before the last character of s, after blanks around it are removed if "t" is included.
var
	string fromWhere,
	int trimEnds,
	int considerCase,
	int pos,
	int l1,
	string sa,
	string sb,
	int changed,
	int changeCount,
	int done

if !stringLength(s)
	return 0
endIf

; Compatibility with older integer whatToDo parameter.
if stringLength(whatToDo) == 1 && stringContains(whatToDo, "0")
	whatToDo = "l"
	elif stringToInt(whatToDo)
	whatToDo = stringSegment("r|a|e|||||lt|rt|at|et|||||lc|rc|ac|ec|||||ltc|rtc|atc|etc", "|", stringToInt(whatToDo) & 31)
endIf

whatToDo = stringLower(whatToDo)
trimEnds = stringContains(whatToDo, "t")
considerCase = stringContains(whatToDo, "c")
fromWhere = whatToDo

; Special cases for null r1.
if !stringLength(r1) && fromWhere == "e"
	if trimEnds
		sa = stringStripAllBlanks(s)
	else
		sa = SkypeWatch__string__ref(s)
	endIf
	s = ""
	l1 = stringLength(sa)
	pos = 1
	while pos < l1
		s = SkypeWatch__string__cat(s, SkypeWatch__string__cat(substring(sa, pos, 1), r2))
		pos = pos +1
	endWhile
	s = SkypeWatch__string__cat(s, stringRight(sa, 1))
	return stringLength(sa)
elif !stringLength(r1)
	if fromWhere == "l" || fromWhere == "a"
		if trimEnds
			sb = stringTrimLeadingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringLeft(sb, 1)
		sb = stringChopLeft(sb, 1)
		if trimEnds
			sb = stringTrimLeadingBlanks(sb)
		endIf
	else  ; fromWhere == "r"
		if trimEnds
			sb = stringTrimTrailingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringChopRight(sb, 1)
		if trimEnds
			sa = stringTrimTrailingBlanks(sa)
		endIf
		sb = stringRight(sb, 1)
	endIf
	if !stringLength(sa) || !stringLength(sb)
		return 0
	endIf
	s = SkypeWatch__string__cat(sa, SkypeWatch__string__cat(r2, sb))
	return 1
endIf

changeCount = 0
while !done
	l1 = stringLength(r1)
	changed = false
	if fromWhere == "l"
		if !stringCompare(stringLeft(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimLeadingBlanks(stringChopLeft(s, l1))
				s = r2 + s
			else
				s = r2 + stringChopLeft(s, l1)
			endIf
			changed = true
		endIf
	elif fromWhere == "r"
		if !stringCompare(stringRight(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimTrailingBlanks(stringChopRight(s, l1))
				s = s + r2
			else
				s = stringChopRight(s, l1) + r2
			endIf
			changed = true
		endIf
	else  ; fromWhere == "a" or "e"
		if considerCase
			pos = StringContains(s, r1)
		else
			pos = StringContains(stringLower(s), stringLower(r1))
		endIf
		if pos
			if trimEnds
				sa = stringTrimTrailingBlanks(stringLeft(s, pos-1))
				sb = stringTrimLeadingBlanks(stringChopLeft(s, pos+l1-1))
				s = " " +sa +r2 +sb
				s = stringChopLeft(s, 1)
			else
				s = " " +stringLeft(s, pos-1) +r2 +stringChopLeft(s, pos+l1-1)
				s = stringChopLeft(s, 1)
			endIf
			changed = True
		endIf
	endIf
	if fromWhere == "e"
		done = !changed
	else
		done = True
	endIf
	if changed
		changeCount = changeCount +1
	endIf
endWhile
return changeCount
endFunction

string function SkypeWatch__string__repeat(string s, int n)
; Return a string consisting of n copies of s, concatenated.
var
	string result
if n <= 0
	return ""
endIf
result = ""
while n > 0
	result = SkypeWatch__string__cat(result, s)
	n = n -1
endWhile
return result
endFunction

string function SkypeWatch__string__ref(variant v)
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

string function SkypeWatch__string__cat(string s1, string s2)
; Safe string concatenator.
return stringChopLeft(" " +s1 +s2, 1)
endFunction

;============================== End module string ==============================

